import unittest

from kafka.tools.protocol.types.integers import Int64


class Int64Tests(unittest.TestCase):
    def test_create(self):
        val = Int64(13147483647)
        assert val.value() == 13147483647

    def test_create_null(self):
        self.assertRaises(TypeError, Int64, None)

    def test_encode_bad_type(self):
        self.assertRaises(TypeError, Int64, 'abc')

    def test_encode_bad_range(self):
        self.assertRaises(ValueError, Int64, 9423372036854775807)

    def test_decode(self):
        (val, rest) = Int64.decode(b'\x00\x00\x00\x03\x0f\xa6\xad\xff')
        assert isinstance(val, Int64)
        assert val.value() == 13147483647
        assert rest == b''

    def test_decode_remainder(self):
        (val, rest) = Int64.decode(b'\x00\x00\x00\x03\x0f\xa6\xad\xffabc')
        assert val.value() == 13147483647
        assert rest == b'abc'

    def test_decode_underflow(self):
        self.assertRaises(ValueError, Int64.decode, b'')

    def test_encode(self):
        val = Int64(13147483647)
        assert val.encode() == b'\x00\x00\x00\x03\x0f\xa6\xad\xff'



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.leave_group_v0 import LeaveGroupV0Request


class LeaveGroupV0RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = LeaveGroupV0Request.process_arguments(['groupname', 'membername'])
        assert val == {'group_id': 'groupname', 'member_id': 'membername'}

    def test_process_arguments_missing(self):
        self.assertRaises(ArgumentError, LeaveGroupV0Request.process_arguments, [])

    def test_process_arguments_three(self):
        self.assertRaises(ArgumentError, LeaveGroupV0Request.process_arguments, ['groupname', 'membername', 'extraarg'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class KafkaToolsException(Exception):
    errstr = "Unknown kafka-tools exception"

    def __init__(self, custom_errstr=None):
        super(KafkaToolsException, self).__init__()
        if custom_errstr is not None:
            self.errstr = custom_errstr

    def __str__(self):
        return self.errstr


class AssignerException(KafkaToolsException):
    errstr = "Unknown Assigner Exception"


class ReplicaNotFoundException(AssignerException):
    errstr = "The specified replica is not present in the partition"


class NotEnoughReplicasException(AssignerException):
    errstr = "There were not enough replicas left for a partition"


class ConfigurationException(AssignerException):
    errstr = "There was an error in the configuration provided"


class ZookeeperException(AssignerException):
    errstr = "There was an error connecting to Zookeeper"


class ClusterConsistencyException(AssignerException):
    errstr = "There is a problem with the consistency of the cluster topics and/or partitions"


class ProgrammingException(AssignerException):
    errstr = "There is an error in the structure of the code"


class ReassignmentFailedException(AssignerException):
    errstr = "The partition reassignment failed"


class UnknownBrokerException(AssignerException):
    errstr = "There is an unknown broker hostname"


class BalanceException(AssignerException):
    errstr = "The balance module failed to satisfy the requirements"


# Errors for the client
class ClientError(KafkaToolsException):
    errstr = "Unknown Client error"


class ConnectionError(ClientError):
    errstr = "There was an error connecting to a Kafka broker"


class GroupError(ClientError):
    errstr = "There was an error getting information about a group"


class TopicError(ClientError):
    errstr = "There was an error getting information about a topic"


class OffsetError(ClientError):
    errstr = "There was an error fetching offsets"


class ConfigurationError(ClientError):
    errstr = "There was an error configuring the client"



import unittest
from mock import MagicMock

from tests.tools.client.fixtures import list_offset, list_offset_error, topic_metadata, offset_commit_response

from kafka.tools.client import Client
from kafka.tools.exceptions import ConnectionError, OffsetError
from kafka.tools.models.topic import TopicOffsets
from kafka.tools.protocol.requests.offset_commit_v2 import OffsetCommitV2Request


class OffsetTests(unittest.TestCase):
    def setUp(self):
        # Dummy client for testing - we're not going to connect that bootstrap broker
        self.client = Client()

        # Get the broker and topic from a metadata update
        self.client._update_from_metadata(topic_metadata())

        self.list_offset = list_offset()
        self.list_offset_error = list_offset_error()
        self.list_offset_request = {1: {'replica_id': -1,
                                        'topics': [{'topic': 'topic1',
                                                    'partitions': [{'partition': 0,
                                                                    'timestamp': Client.OFFSET_LATEST,
                                                                    'max_num_offsets': 1},
                                                                   {'partition': 1,
                                                                    'timestamp': Client.OFFSET_EARLIEST,
                                                                    'max_num_offsets': 1}]}]}}

    def test_send_list_offsets_to_brokers(self):
        self.client.cluster.brokers[1].send = MagicMock()
        self.client.cluster.brokers[1].send.return_value = (1, self.list_offset)

        val = self.client._send_list_offsets_to_brokers(self.list_offset_request)

        self.client.cluster.brokers[1].send.assert_called_once()
        req = self.client.cluster.brokers[1].send.call_args[0][0]
        assert req['replica_id'] == -1
        assert len(req['topics']) == 1
        assert req['topics'][0]['topic'].value() == 'topic1'
        assert len(req['topics'][0]['partitions']) == 2
        assert req['topics'][0]['partitions'][0]['partition'] == 0
        assert req['topics'][0]['partitions'][0]['timestamp'] == Client.OFFSET_LATEST
        assert req['topics'][0]['partitions'][0]['max_num_offsets'] == 1
        assert req['topics'][0]['partitions'][1]['partition'] == 1
        assert req['topics'][0]['partitions'][1]['timestamp'] == Client.OFFSET_EARLIEST
        assert req['topics'][0]['partitions'][1]['max_num_offsets'] == 1

        assert len(val) == 1
        assert 'topic1' in val
        assert isinstance(val['topic1'], TopicOffsets)
        assert val['topic1'].topic == self.client.cluster.topics['topic1']
        assert len(val['topic1'].partitions) == 2
        assert val['topic1'].partitions[0] == 4829
        assert val['topic1'].partitions[1] == 8904

    def test_send_list_offsets_to_brokers_connection_error(self):
        self.client.cluster.brokers[1].send = MagicMock()
        self.client.cluster.brokers[1].send.side_effect = ConnectionError
        self.assertRaises(ConnectionError, self.client._send_list_offsets_to_brokers, self.list_offset_request)

    def test_send_list_offsets_to_brokers_offsets_error(self):
        self.client.cluster.brokers[1].send = MagicMock()
        self.client.cluster.brokers[1].send.return_value = (1, self.list_offset_error)
        self.assertRaises(OffsetError, self.client._send_list_offsets_to_brokers, self.list_offset_request)

    def test_send_set_offset_request_bad_offsets(self):
        self.assertRaises(TypeError, self.client._send_set_offset_request, 'testgroup', ['notatopicoffsets'])

    def test_send_set_offset_request_bad_topic(self):
        offsets = TopicOffsets(self.client.cluster.topics['topic1'])
        offsets.topic = None
        self.assertRaises(TypeError, self.client._send_set_offset_request, 'testgroup', [offsets])

    def test_send_set_offset_request(self):
        offsets = TopicOffsets(self.client.cluster.topics['topic1'])
        offsets.partitions[0] = 2342
        offsets.partitions[1] = 8793

        self.client._send_group_aware_request = MagicMock()
        self.client._send_group_aware_request.return_value = 'responseobject'
        val = self.client._send_set_offset_request('testgroup', [offsets])
        assert val == 'responseobject'

        self.client._send_group_aware_request.assert_called_once()
        assert self.client._send_group_aware_request.call_args[0][0] == 'testgroup'
        req = self.client._send_group_aware_request.call_args[0][1]
        assert isinstance(req, OffsetCommitV2Request)
        assert req['group_id'].value() == 'testgroup'
        assert req['group_generation_id'].value() == -1
        assert req['member_id'].value() == ''
        assert req['retention_time'].value() == -1
        assert len(req['topics']) == 1
        assert req['topics'][0]['topic'].value() == 'topic1'
        assert len(req['topics'][0]['partitions']) == 2
        assert req['topics'][0]['partitions'][0]['partition'].value() == 0
        assert req['topics'][0]['partitions'][0]['offset'].value() == 2342
        assert req['topics'][0]['partitions'][0]['metadata'].value() is None
        assert req['topics'][0]['partitions'][1]['partition'].value() == 1
        assert req['topics'][0]['partitions'][1]['offset'].value() == 8793
        assert req['topics'][0]['partitions'][1]['metadata'].value() is None

    def test_parse_set_offset_response(self):
        response = offset_commit_response()
        val = self.client._parse_set_offset_response(response)
        assert val == {'topic1': [0, 16]}



import unittest
from mock import patch

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.create_topics_v0 import _parse_partition, _parse_kv_args, _parse_remaining_args, CreateTopicsV0Request


class CreateTopicsV0RequestTests(unittest.TestCase):
    def setUp(self):
        self.topic = {'topic': 'topicname',
                      'num_partitions': -1,
                      'replication_factor': -1,
                      'replica_assignment': [],
                      'configs': []}
        self.target_topic = {'topic': 'topicname',
                             'num_partitions': -1,
                             'replication_factor': -1,
                             'replica_assignment': [],
                             'configs': []}

    def test_parse_partition(self):
        _parse_partition(self.topic, 1, "3|4")

        self.target_topic['replica_assignment'] = [{'partition_id': 1, 'replicas': [3, 4]}]
        assert self.topic == self.target_topic

    def test_parse_partition_conflict(self):
        self.topic['num_partitions'] = 8
        self.topic['replication_factor'] = 3
        self.assertRaises(ArgumentError, _parse_partition, self.topic, 1, "3|4")

    def test_parse_partition_nonnumeric(self):
        self.assertRaises(ArgumentError, _parse_partition, self.topic, 1, "3|foo")

    def test_parse_kv_args_partition(self):
        _parse_kv_args(self.topic, ["1=3|4"])

        self.target_topic['replica_assignment'] = [{'partition_id': 1, 'replicas': [3, 4]}]
        assert self.topic == self.target_topic

    def test_parse_kv_args_config(self):
        _parse_kv_args(self.topic, ["configkey=configvalue"])

        self.target_topic['configs'] = [{'config_key': 'configkey', 'config_value': 'configvalue'}]
        assert self.topic == self.target_topic

    def test_parse_kv_args_noequals(self):
        self.topic['num_partitions'] = 8
        self.assertRaises(ArgumentError, _parse_kv_args, self.topic, "foo")

    def test_parse_remaining_args_count_and_factor(self):
        _parse_remaining_args(self.topic, ["8", "3", "configkey=configvalue"])

        self.target_topic['num_partitions'] = 8
        self.target_topic['replication_factor'] = 3
        self.target_topic['configs'] = [{'config_key': 'configkey', 'config_value': 'configvalue'}]
        assert self.topic == self.target_topic

    def test_parse_remaining_args_replica_list(self):
        _parse_remaining_args(self.topic, ["1=3|4", "configkey=configvalue"])

        self.target_topic['replica_assignment'] = [{'partition_id': 1, 'replicas': [3, 4]}]
        self.target_topic['configs'] = [{'config_key': 'configkey', 'config_value': 'configvalue'}]
        assert self.topic == self.target_topic

    def test_parse_remaining_args_no_partitions(self):
        self.assertRaises(ArgumentError, _parse_remaining_args, self.topic, ["configkey=configvalue"])

    @patch('kafka.tools.protocol.requests.create_topics_v0._parse_remaining_args')
    def test_process_arguments(self, mock_parse):
        val = CreateTopicsV0Request.process_arguments(['3296', 'topicname,someargs'])
        mock_parse.assert_called_once_with(self.topic, ['someargs'])
        assert val == {'timeout': 3296, 'topics': [self.topic]}

    def test_process_arguments_notimeout(self):
        self.assertRaises(ArgumentError, CreateTopicsV0Request.process_arguments, ['topicname,someargs', 'anothertopic,moreargs'])

    def test_process_arguments_notopic(self):
        self.assertRaises(ArgumentError, CreateTopicsV0Request.process_arguments, ['3296'])



import sys
import unittest

from argparse import Namespace
from collections import deque
from mock import call, patch
from ..fixtures import set_up_cluster, set_up_subparser

from kafka.tools.exceptions import BalanceException
from kafka.tools.models.broker import Broker
from kafka.tools.assigner.actions.balance import ActionBalance
from kafka.tools.assigner.actions.balancemodules.rackaware import (ActionBalanceRackAware, check_partition_swappable, racks_for_replica_list,
                                                                   difference_in_size_to_last_partition)


class ActionBalanceRackAwareTests(unittest.TestCase):
    def setUp(self):
        self.cluster = set_up_cluster()
        self.cluster.topics['testTopic1'].partitions[0].size = 1000
        self.cluster.topics['testTopic1'].partitions[1].size = 1000
        self.cluster.topics['testTopic2'].partitions[0].size = 2000
        self.cluster.topics['testTopic2'].partitions[1].size = 2000

        (self.parser, self.subparsers) = set_up_subparser()
        self.args = Namespace(exclude_topics=[])

    def test_configure_args(self):
        ActionBalance.configure_args(self.subparsers)
        sys.argv = ['kafka-assigner', 'balance', '-t', 'rackaware']
        parsed_args = self.parser.parse_args()
        assert parsed_args.action == 'balance'

    def test_init_broker_deque(self):
        action = ActionBalanceRackAware(self.args, self.cluster)
        assert set(action._random_brokers) == set(self.cluster.brokers.values())

    def test_check_partition_swappable_already_exists(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        replicas_a = [b1, b2]
        replicas_b = [b2, b1]
        assert check_partition_swappable(replicas_a, replicas_b, 0) is False
        assert check_partition_swappable(replicas_b, replicas_a, 0) is False

    def test_check_partition_swappable_racks_collide(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        b3 = Broker('brokerhost3.example.com', id=3)
        b4 = Broker('brokerhost4.example.com', id=4)
        b2.rack = "a"
        b3.rack = "a"
        b4.rack = "b"
        replicas_a = [b1, b2]
        replicas_b = [b3, b4]
        assert check_partition_swappable(replicas_a, replicas_b, 0) is False
        assert check_partition_swappable(replicas_b, replicas_a, 0) is False

    def test_check_partition_swappable_racks_ok(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        b3 = Broker('brokerhost3.example.com', id=3)
        b4 = Broker('brokerhost4.example.com', id=4)
        b2.rack = "a"
        b3.rack = "b"
        b4.rack = "b"
        replicas_a = [b1, b2]
        replicas_b = [b3, b4]
        assert check_partition_swappable(replicas_a, replicas_b, 0) is True
        assert check_partition_swappable(replicas_b, replicas_a, 0) is True

    def test_racks_for_replica_list_nopos(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        assert racks_for_replica_list([b1, b2]) == ["a", "b"]

    def test_racks_for_replica_list_pos(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        assert racks_for_replica_list([b1, b2], 1) == ["a"]

    def test_difference_in_size_larger(self):
        partitions = [self.cluster.topics['testTopic1'].partitions[0],
                      self.cluster.topics['testTopic1'].partitions[1]]
        partitions[0].size = 1000
        partitions[1].size = 2000
        p = self.cluster.topics['testTopic2'].partitions[0]
        p.size = 3000
        assert difference_in_size_to_last_partition(p, partitions) == 1000

    def test_difference_in_size_smaller(self):
        partitions = [self.cluster.topics['testTopic1'].partitions[0],
                      self.cluster.topics['testTopic1'].partitions[1]]
        partitions[0].size = 1000
        partitions[1].size = 2000
        p = self.cluster.topics['testTopic2'].partitions[0]
        p.size = 1000
        assert difference_in_size_to_last_partition(p, partitions) == 1000

    def test_difference_in_size_empty_list(self):
        partitions = []
        p = self.cluster.topics['testTopic2'].partitions[0]
        p.size = 1000
        assert difference_in_size_to_last_partition(p, partitions) == float("inf")

    def test_try_pick_new_broker_skipself(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        b3 = Broker('brokerhost3.example.com', id=3)
        self.cluster.add_broker(b3)
        b3.rack = "a"
        action = ActionBalanceRackAware(self.args, self.cluster)

        # Firmly order the deque
        action._random_brokers = deque([b1, b2, b3])
        newbroker = action._try_pick_new_broker(self.cluster.topics['testTopic1'].partitions[0], 0)
        assert newbroker == b3
        assert action._random_brokers == deque([b1, b2, b3])

    def test_try_pick_new_broker_failed(self):
        action = ActionBalanceRackAware(self.args, self.cluster)
        self.assertRaises(BalanceException, action._try_pick_new_broker, self.cluster.topics['testTopic1'].partitions[0], 0)

    def test_try_pick_new_broker(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        b3 = Broker('brokerhost3.example.com', id=3)
        self.cluster.add_broker(b3)
        b3.rack = "b"
        action = ActionBalanceRackAware(self.args, self.cluster)

        # Firmly order the deque
        action._random_brokers = deque([b3, b1, b2])
        newbroker = action._try_pick_new_broker(self.cluster.topics['testTopic1'].partitions[0], 1)
        assert newbroker == b3
        assert action._random_brokers == deque([b1, b2, b3])

    def test_try_pick_swap_partition_none(self):
        action = ActionBalanceRackAware(self.args, self.cluster)
        small_partitions = [self.cluster.topics['testTopic1'].partitions[0],
                            self.cluster.topics['testTopic1'].partitions[1]]
        large_partitions = [self.cluster.topics['testTopic2'].partitions[0]]
        p = self.cluster.topics['testTopic2'].partitions[1]
        small_partitions[0].size = 1000
        small_partitions[1].size = 2000
        large_partitions[0].size = 4000
        p.size = 3000
        assert action._try_pick_swap_partition(p, 0, small_partitions, large_partitions) is None

    @patch('kafka.tools.assigner.actions.balancemodules.rackaware.check_partition_swappable')
    def test_try_pick_swap_partitions_nosmall(self, mock_check):
        action = ActionBalanceRackAware(self.args, self.cluster)
        small_partitions = []
        large_partitions = [self.cluster.topics['testTopic2'].partitions[0]]
        p = self.cluster.topics['testTopic2'].partitions[1]
        large_partitions[0].size = 4000
        p.size = 3000

        mock_check.return_value = True
        assert action._try_pick_swap_partition(p, 0, small_partitions, large_partitions) == self.cluster.topics['testTopic2'].partitions[0]

    @patch('kafka.tools.assigner.actions.balancemodules.rackaware.check_partition_swappable')
    def test_try_pick_swap_partitions_nolarge(self, mock_check):
        action = ActionBalanceRackAware(self.args, self.cluster)
        large_partitions = []
        small_partitions = [self.cluster.topics['testTopic2'].partitions[0]]
        p = self.cluster.topics['testTopic2'].partitions[1]
        small_partitions[0].size = 1000
        p.size = 3000

        mock_check.return_value = True
        assert action._try_pick_swap_partition(p, 0, small_partitions, large_partitions) == self.cluster.topics['testTopic2'].partitions[0]

    @patch('kafka.tools.assigner.actions.balancemodules.rackaware.check_partition_swappable')
    def test_try_pick_swap_partition_full(self, mock_check):
        action = ActionBalanceRackAware(self.args, self.cluster)
        small_partitions = [self.cluster.topics['testTopic1'].partitions[0],
                            self.cluster.topics['testTopic1'].partitions[1]]
        large_partitions = [self.cluster.topics['testTopic2'].partitions[0]]
        p = self.cluster.topics['testTopic2'].partitions[1]
        small_partitions[0].size = 1000
        small_partitions[1].size = 2000
        large_partitions[0].size = 4000
        p.size = 3000

        mock_check.side_effect = [False, False, True]
        target = action._try_pick_swap_partition(p, 0, small_partitions, large_partitions)

        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        mock_check.assert_has_calls([call([b1, b2], [b2, b1], 0), call([b1, b2], [b2, b1], 0), call([b1, b2], [b1, b2], 0)])
        assert target == self.cluster.topics['testTopic1'].partitions[0]

    def test_get_sorted_list(self):
        action = ActionBalanceRackAware(self.args, self.cluster)
        p1 = self.cluster.topics['testTopic1'].partitions[0]
        p2 = self.cluster.topics['testTopic1'].partitions[1]
        p3 = self.cluster.topics['testTopic2'].partitions[0]
        p4 = self.cluster.topics['testTopic2'].partitions[1]
        p1.size = 1000
        p2.size = 6000
        p3.size = 3000
        p4.size = 4000
        assert action._get_sorted_partition_list_at_pos(0) == [p1, p3, p4, p2]

    def test_get_sorted_list_missing_replica(self):
        action = ActionBalanceRackAware(self.args, self.cluster)
        p1 = self.cluster.topics['testTopic1'].partitions[0]
        p2 = self.cluster.topics['testTopic1'].partitions[1]
        p3 = self.cluster.topics['testTopic2'].partitions[0]
        p4 = self.cluster.topics['testTopic2'].partitions[1]
        p1.size = 1000
        p2.size = 6000
        p3.size = 3000
        p4.size = 4000
        p1.replicas = [self.cluster.brokers[1]]
        assert action._get_sorted_partition_list_at_pos(0) == [p3, p4, p2]

    @patch.object(ActionBalanceRackAware, '_try_pick_swap_partition')
    def test_process_partitions_at_pos_nochange(self, mock_pick):
        action = ActionBalanceRackAware(self.args, self.cluster)
        action._process_partitions_at_pos(0)
        mock_pick.assert_not_called()

    @patch.object(ActionBalanceRackAware, '_try_pick_swap_partition')
    def test_process_partitions_at_pos_swap_partition(self, mock_pick):
        action = ActionBalanceRackAware(self.args, self.cluster)
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        b3 = Broker('brokerhost3.example.com', id=3)
        b4 = Broker('brokerhost4.example.com', id=4)
        self.cluster.add_broker(b3)
        self.cluster.add_broker(b4)
        b3.rack = "a"
        b4.rack = "c"
        self.cluster.topics['testTopic2'].partitions[0].swap_replicas(b2, b3)
        self.cluster.topics['testTopic1'].partitions[1].swap_replicas(b1, b4)
        mock_pick.return_value = self.cluster.topics['testTopic1'].partitions[1]

        action._process_partitions_at_pos(0)
        assert self.cluster.topics['testTopic1'].partitions[1].replicas == [b3, b4]
        assert self.cluster.topics['testTopic2'].partitions[0].replicas == [b2, b1]

    @patch.object(ActionBalanceRackAware, '_try_pick_swap_partition')
    @patch.object(ActionBalanceRackAware, '_try_pick_new_broker')
    def test_process_partitions_at_pos_swap_broker(self, mock_broker, mock_pick):
        action = ActionBalanceRackAware(self.args, self.cluster)
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        b3 = Broker('brokerhost3.example.com', id=3)
        b4 = Broker('brokerhost4.example.com', id=4)
        self.cluster.add_broker(b3)
        self.cluster.add_broker(b4)
        b3.rack = "a"
        b4.rack = "c"
        self.cluster.topics['testTopic2'].partitions[0].swap_replicas(b2, b3)
        self.cluster.topics['testTopic1'].partitions[1].swap_replicas(b1, b4)
        mock_pick.return_value = None
        mock_broker.return_value = b2

        action._process_partitions_at_pos(0)
        assert self.cluster.topics['testTopic1'].partitions[1].replicas == [b2, b4]
        assert self.cluster.topics['testTopic2'].partitions[0].replicas == [b2, b1]

    def test_process_cluster_single_rack(self):
        action = ActionBalanceRackAware(self.args, self.cluster)
        self.cluster.brokers[2].rack = "a"
        self.assertRaises(BalanceException, action.process_cluster)

    @patch.object(ActionBalanceRackAware, '_process_partitions_at_pos')
    def test_process_cluster(self, mock_process):
        action = ActionBalanceRackAware(self.args, self.cluster)
        action.process_cluster()
        mock_process.assert_has_calls([call(0), call(1)])



import time
import unittest
from mock import MagicMock

from tests.tools.client.fixtures import topic_metadata

from kafka.tools.client import Client
from kafka.tools.exceptions import TopicError
from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.protocol.requests.topic_metadata_v1 import TopicMetadataV1Request


def assert_cluster_has_topics(cluster, metadata):
    for mtopic in metadata['topics']:
        assert mtopic['name'].value() in cluster.topics
        topic = cluster.topics[mtopic['name'].value()]
        assert topic.name == mtopic['name'].value()
        assert topic.internal == mtopic['internal'].value()
        assert len(topic.partitions) == len(mtopic['partitions'])

        for i, tp in enumerate(mtopic['partitions']):
            partition = topic.partitions[i]
            assert partition.num == tp['id'].value()
            assert partition.leader.id == tp['leader'].value()
            assert len(partition.replicas) == len(tp['replicas'])

            for j, tp_replica in enumerate(tp['replicas']):
                assert partition.replicas[j].id == tp_replica


def assert_cluster_has_brokers(cluster, metadata):
    for b in metadata['brokers']:
        assert b['node_id'].value() in cluster.brokers
        broker = cluster.brokers[b['node_id'].value()]
        assert broker.hostname == b['host'].value()
        assert broker.port == b['port'].value()
        assert broker.rack == b['rack'].value()


class TopicsTests(unittest.TestCase):
    def setUp(self):
        # Dummy client for testing - we're not going to connect that bootstrap broker
        self.client = Client()
        self.metadata_response = topic_metadata()

    def test_maybe_update_full_metadata_expired(self):
        self.client._send_any_broker = MagicMock()
        self.client._send_any_broker.return_value = 'metadata_response'
        self.client._update_from_metadata = MagicMock()

        fake_last_time = time.time() - (self.client.configuration.metadata_refresh * 2)
        self.client._last_full_metadata = fake_last_time
        self.client._maybe_update_full_metadata()

        assert self.client._last_full_metadata > fake_last_time
        self.client._send_any_broker.assert_called_once()
        arg = self.client._send_any_broker.call_args[0][0]
        assert isinstance(arg, TopicMetadataV1Request)
        assert arg['topics'].value() is None
        self.client._update_from_metadata.assert_called_once_with('metadata_response', delete=True)

    def test_maybe_update_full_metadata_nocache(self):
        self.client._send_any_broker = MagicMock()
        self.client._update_from_metadata = MagicMock()

        fake_last_time = time.time() - 1000
        self.client._last_full_metadata = fake_last_time
        self.client._maybe_update_full_metadata(cache=False)

        assert self.client._last_full_metadata > fake_last_time

    def test_maybe_update_full_metadata_usecache(self):
        self.client._send_any_broker = MagicMock()
        self.client._update_from_metadata = MagicMock()

        fake_last_time = time.time() - 1000
        self.client._last_full_metadata = fake_last_time
        self.client._maybe_update_full_metadata(cache=True)

        assert self.client._last_full_metadata == fake_last_time

    def test_maybe_update_metadata_for_topics_noupdate(self):
        self.client._update_from_metadata(self.metadata_response)
        self.client.cluster.topics['topic1']._last_updated = time.time()

        self.client._update_from_metadata = MagicMock()
        self.client._send_any_broker = MagicMock()
        self.client._maybe_update_metadata_for_topics(['topic1'])
        self.client._update_from_metadata.assert_not_called()

    def test_maybe_update_metadata_for_topics_expired(self):
        self.client._update_from_metadata(self.metadata_response)
        self.client.cluster.topics['topic1']._last_updated = 100

        self.client._update_from_metadata = MagicMock()
        self.client._send_any_broker = MagicMock()
        self.client._maybe_update_metadata_for_topics(['topic1'])
        self.client._update_from_metadata.assert_called_once()
        req = self.client._send_any_broker.call_args[0][0]
        assert len(req['topics']) == 1
        assert req['topics'][0].value() == 'topic1'

    def test_maybe_update_metadata_for_topics_forced(self):
        self.client._update_from_metadata(self.metadata_response)
        self.client.cluster.topics['topic1']._last_updated = time.time()

        self.client._update_from_metadata = MagicMock()
        self.client._send_any_broker = MagicMock()
        self.client._maybe_update_metadata_for_topics(['topic1'], cache=False)
        self.client._update_from_metadata.assert_called_once()
        req = self.client._send_any_broker.call_args[0][0]
        assert len(req['topics']) == 1
        assert req['topics'][0].value() == 'topic1'

    def test_maybe_update_metadata_for_topics_nonexistent(self):
        self.client._update_from_metadata(self.metadata_response)
        self.client.cluster.topics['topic1']._last_updated = time.time()

        self.client._update_from_metadata = MagicMock()
        self.client._send_any_broker = MagicMock()
        self.client._maybe_update_metadata_for_topics(['topic1', 'topic2'])
        self.client._update_from_metadata.assert_called_once()
        req = self.client._send_any_broker.call_args[0][0]
        assert len(req['topics']) == 2
        assert req['topics'][0].value() == 'topic1'
        assert req['topics'][1].value() == 'topic2'

    def test_update_from_metadata(self):
        self.client._update_brokers_from_metadata = MagicMock()
        self.client._update_topics_from_metadata = MagicMock()

        self.client._update_from_metadata('fake_metadata')
        self.client._update_brokers_from_metadata.assert_called_once_with('fake_metadata')
        self.client._update_topics_from_metadata.assert_called_once_with('fake_metadata', delete=False)

    def test_update_topics_from_metadata_create(self):
        # Don't want to test the broker update code here
        self.client.cluster.add_broker(Broker('host1.example.com', id=1, port=8031))
        self.client.cluster.add_broker(Broker('host2.example.com', id=101, port=8032))

        self.client._update_topics_from_metadata(self.metadata_response)
        assert_cluster_has_topics(self.client.cluster, self.metadata_response)

    def test_maybe_delete_topics_not_in_metadata(self):
        # Don't want to test the broker update code here
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        topic = Topic('topic2', 1)
        self.client.cluster.add_broker(broker1)
        self.client.cluster.add_broker(broker2)
        self.client.cluster.add_topic(topic)
        topic.partitions[0].add_replica(broker2)
        topic.partitions[0].add_replica(broker1)

        self.client._maybe_delete_topics_not_in_metadata(self.metadata_response, delete=True)
        assert 'topic2' not in self.client.cluster.topics

    def test_update_topics_from_metadata_update_replicas(self):
        # Don't want to test the broker update code here
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        broker3 = Broker('host3.example.com', id=304, port=8033)
        topic = Topic('topic1', 2)
        self.client.cluster.add_broker(broker1)
        self.client.cluster.add_broker(broker2)
        self.client.cluster.add_broker(broker3)
        self.client.cluster.add_topic(topic)
        topic.partitions[0].add_replica(broker3)
        topic.partitions[0].add_replica(broker1)
        topic.partitions[1].add_replica(broker2)
        topic.partitions[1].add_replica(broker1)

        self.client._update_topics_from_metadata(self.metadata_response)
        assert_cluster_has_topics(self.client.cluster, self.metadata_response)

    def test_update_topics_from_metadata_delete_replicas(self):
        # Don't want to test the broker update code here
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        broker3 = Broker('host3.example.com', id=304, port=8033)
        topic = Topic('topic1', 2)
        self.client.cluster.add_broker(broker1)
        self.client.cluster.add_broker(broker2)
        self.client.cluster.add_broker(broker3)
        self.client.cluster.add_topic(topic)
        topic.partitions[0].add_replica(broker2)
        topic.partitions[0].add_replica(broker1)
        topic.partitions[0].add_replica(broker3)
        topic.partitions[1].add_replica(broker2)
        topic.partitions[1].add_replica(broker1)

        self.client._update_topics_from_metadata(self.metadata_response)
        assert_cluster_has_topics(self.client.cluster, self.metadata_response)

    def test_update_brokers_from_metadata(self):
        self.client._update_brokers_from_metadata(self.metadata_response)
        assert_cluster_has_brokers(self.client.cluster, self.metadata_response)

    def test_update_brokers_from_metadata_update_rack(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker1.rack = 'wrongrack'
        self.client.cluster.add_broker(broker1)

        self.client._update_brokers_from_metadata(self.metadata_response)
        assert_cluster_has_brokers(self.client.cluster, self.metadata_response)

    def test_update_brokers_from_metadata_update_host(self):
        broker1 = Broker('wronghost.example.com', id=1, port=8031)
        self.client.cluster.add_broker(broker1)
        broker1.close = MagicMock()

        self.client._update_brokers_from_metadata(self.metadata_response)
        assert_cluster_has_brokers(self.client.cluster, self.metadata_response)
        broker1.close.assert_called_once()

    def test_map_topic_partitions_to_brokers(self):
        self.client._update_from_metadata(self.metadata_response)
        val = self.client._map_topic_partitions_to_brokers(['topic1'])
        assert val == {1: {'topic1': [0]}, 101: {'topic1': [1]}}

    def test_map_topic_partitions_to_brokers_nonexistent(self):
        self.client._update_from_metadata(self.metadata_response)
        self.assertRaises(TopicError, self.client._map_topic_partitions_to_brokers, ['nosuchtopic'])



import time
import unittest
from mock import MagicMock

from tests.tools.client.fixtures import describe_groups, list_groups, list_groups_error

from kafka.tools.client import Client
from kafka.tools.models.broker import Broker
from kafka.tools.models.group import Group, GroupMember
from kafka.tools.protocol.requests.list_groups_v0 import ListGroupsV0Request


def assert_cluster_has_groups(cluster, dg):
    for dgroup in dg['groups']:
        assert dgroup['group_id'].value() in cluster.groups
        group = cluster.groups[dgroup['group_id'].value()]
        assert group.name == dgroup['group_id'].value()
        assert group.protocol == dgroup['protocol'].value()
        assert group.protocol_type == dgroup['protocol_type'].value()
        assert len(group.members) == len(dgroup['members'])

        for i, gmember in enumerate(dgroup['members']):
            member = group.members[i]
            assert member.group == group
            assert member.name == gmember['member_id'].value()
            assert member.client_id == gmember['client_id'].value()
            assert member.client_host == gmember['client_host'].value()
            assert member.metadata == gmember['member_metadata'].value()
            assert member.assignment_data == gmember['member_assignment'].value()


class GroupsTests(unittest.TestCase):
    def setUp(self):
        # Dummy client for testing - we're not going to connect that bootstrap broker
        self.client = Client()
        self.describe_groups = describe_groups()

    def test_update_groups_from_describe_create(self):
        self.client._update_groups_from_describe(self.describe_groups)
        assert_cluster_has_groups(self.client.cluster, self.describe_groups)

    def test_update_groups_from_describe_assignment(self):
        self.client._update_groups_from_describe(self.describe_groups)
        assert_cluster_has_groups(self.client.cluster, self.describe_groups)

        group = self.client.cluster.groups['testgroup']
        assert group.members[0].topics == {'topic1': [0]}
        assert group.members[1].topics == {'topic1': [1]}

    def test_update_groups_from_describe_update(self):
        self.client.cluster.add_group(Group('testgroup'))
        self.client._update_groups_from_describe(self.describe_groups)

        assert_cluster_has_groups(self.client.cluster, self.describe_groups)

    def test_update_groups_from_describe_clear_members(self):
        self.client.cluster.add_group(Group('testgroup'))
        self.client.cluster.groups['testgroup'].members = [GroupMember('badmember')]
        self.client._update_groups_from_describe(self.describe_groups)

        assert_cluster_has_groups(self.client.cluster, self.describe_groups)

    def test_maybe_update_groups_list_expired(self):
        self.client._send_all_brokers = MagicMock()
        self.client._send_all_brokers.return_value = ['metadata_response']
        self.client._update_groups_from_lists = MagicMock()

        fake_last_time = time.time() - (self.client.configuration.metadata_refresh * 2)
        self.client._last_group_list = fake_last_time
        self.client._maybe_update_groups_list()

        assert self.client._last_group_list > fake_last_time
        self.client._send_all_brokers.assert_called_once()
        arg = self.client._send_all_brokers.call_args[0][0]
        assert isinstance(arg, ListGroupsV0Request)
        self.client._update_groups_from_lists.assert_called_once_with(['metadata_response'])

    def test_maybe_update_groups_list_nocache(self):
        self.client._send_all_brokers = MagicMock()
        self.client._update_groups_from_lists = MagicMock()

        fake_last_time = time.time() - 1000
        self.client._last_group_list = fake_last_time
        self.client._maybe_update_groups_list(cache=False)

        assert self.client._last_group_list > fake_last_time

    def test_maybe_update_groups_list_usecache(self):
        self.client._send_all_brokers = MagicMock()
        self.client._update_groups_from_lists = MagicMock()

        fake_last_time = time.time() - 1000
        self.client._last_group_list = fake_last_time
        self.client._maybe_update_groups_list(cache=True)

    def test_update_groups_from_lists(self):
        self.client._add_or_update_group = MagicMock()
        list_group = list_groups()
        val = self.client._update_groups_from_lists({1: list_group})

        assert val == 0
        self.client._add_or_update_group.assert_called_once_with(list_group['groups'][0], 1)

    def test_update_groups_from_lists_error(self):
        self.client._add_or_update_group = MagicMock()
        val = self.client._update_groups_from_lists({1: list_groups_error()})

        assert val == 1
        self.client._add_or_update_group.assert_not_called()

    def test_update_groups_from_lists_none(self):
        self.client._add_or_update_group = MagicMock()
        val = self.client._update_groups_from_lists({1: None})

        assert val == 1
        self.client._add_or_update_group.assert_not_called()

    def test_add_or_update_group(self):
        broker = Broker('host1.example.com', id=1, port=8031)
        self.client.cluster.add_broker(broker)
        list_group = list_groups()

        self.client._add_or_update_group(list_group['groups'][0], 1)
        assert 'group1' in self.client.cluster.groups
        assert self.client.cluster.groups['group1'].coordinator == broker
        assert self.client.cluster.groups['group1'].protocol_type == 'protocol1'

    def test_add_or_update_group_update(self):
        broker = Broker('host1.example.com', id=1, port=8031)
        self.client.cluster.add_broker(broker)
        group = Group('group1')
        self.client.cluster.add_group(group)
        list_group = list_groups()

        self.client._add_or_update_group(list_group['groups'][0], 1)
        assert 'group1' in self.client.cluster.groups
        assert self.client.cluster.groups['group1'].coordinator == broker
        assert self.client.cluster.groups['group1'].protocol_type == 'protocol1'



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class BaseType():  # pragma: no cover
    @abc.abstractmethod
    def decode(cls, byte_array, schema=None):
        """IMPORTANT: this is class method, override it with @classmethod!"""
        raise NotImplementedError

    def __init__(self, val, schema=None, skip_validation=False):
        self._value = val
        self._schema = schema
        if not skip_validation:
            self._validate()

    def value(self):
        return self._value

    @abc.abstractmethod
    def encode(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _validate(self, val):
        raise NotImplementedError



import argparse
import timeout_decorator
import unittest
from mock import patch, MagicMock

from kafka.tools.models.broker import Broker
from kafka.tools.protocol.__main__ import _get_request_classes, _get_request_commands, _print_errors, _parse_command, _cli_loop, main


class TestKlass1:
    cmd = "TestKlass"
    api_version = 1


class TestKlass2:
    cmd = "TestKlass"
    api_version = 2


class TestKlassy1:
    cmd = "TestKlassy"
    api_version = 1


class MainTests(unittest.TestCase):
    def setUp(self):
        self.broker = Broker('testhost', port=3945)
        self.request_classes = {'testklass': {1: TestKlass1, 2: TestKlass2}, 'testklassy': {1: TestKlassy1}}
        self.request_cmds = {'testklass': TestKlass2,
                             'testklassv1': TestKlass1,
                             'testklassv2': TestKlass2,
                             'testklassy': TestKlassy1,
                             'testklassyv1': TestKlassy1}

    @patch('kafka.tools.protocol.__main__.get_modules')
    def test_get_request_classes(self, mock_modules):
        mock_modules.return_value = [TestKlass1, TestKlass2, TestKlassy1]
        val = _get_request_classes()
        assert val == self.request_classes

    def test_get_request_commands(self):
        val = _get_request_commands({'testklass': {1: TestKlass1, 2: TestKlass2}, 'testklassy': {1: TestKlassy1}})
        print(val)
        assert val == self.request_cmds

    def test_print_errors(self):
        # This just outputs to the console, so just make sure it doesn't fail
        _print_errors()
        assert True

    def test_parse_command_exit(self):
        self.assertRaises(EOFError, _parse_command, self.broker, self.request_classes, self.request_cmds, 'exit', [])
        self.assertRaises(EOFError, _parse_command, self.broker, self.request_classes, self.request_cmds, 'quit', [])
        self.assertRaises(EOFError, _parse_command, self.broker, self.request_classes, self.request_cmds, 'q', [])

    @patch('kafka.tools.protocol.__main__.show_help')
    def test_parse_command_help(self, mock_help):
        _parse_command(self.broker, self.request_classes, self.request_cmds, 'help', [])
        mock_help.assert_called_once_with(self.request_classes, self.request_cmds, [])

    @patch('kafka.tools.protocol.__main__.show_help')
    def test_parse_command_help_pass_args(self, mock_help):
        _parse_command(self.broker, self.request_classes, self.request_cmds, 'help', ['foo'])
        mock_help.assert_called_once_with(self.request_classes, self.request_cmds, ['foo'])

    @patch('kafka.tools.protocol.__main__._print_errors')
    def test_parse_command_errors(self, mock_errors):
        _parse_command(self.broker, self.request_classes, self.request_cmds, 'errors', [])
        mock_errors.assert_called_once_with()

    def test_parse_command_unknown(self):
        # Should just not raise an error, as it prints out
        _parse_command(self.broker, self.request_classes, self.request_cmds, 'unknown_command', [])
        assert True

    def test_parse_command_request(self):
        mock_klass = MagicMock()
        mock_broker = MagicMock()
        self.request_cmds['testklass'] = mock_klass
        self.broker = mock_broker

        mock_klass.process_arguments.return_value = 'fake_request_dict'
        mock_klass.return_value = 'fake_request'
        mock_broker.send.return_value = (8129, 'fake_response')
        _parse_command(self.broker, self.request_classes, self.request_cmds, 'testklass', ['someargs'])

        mock_klass.process_arguments.assert_called_once_with(['someargs'])
        mock_klass.assert_called_once_with('fake_request_dict')
        mock_broker.send.assert_called_once_with('fake_request')

    @patch('kafka.tools.protocol.__main__._get_request_classes')
    @patch('kafka.tools.protocol.__main__._get_request_commands')
    @patch('kafka.tools.protocol.__main__._parse_command')
    @patch('kafka.tools.protocol.__main__.input')
    @timeout_decorator.timeout(5)
    def test_cli_loop_command(self, mock_input, mock_parse, mock_commands, mock_classes):
        mock_classes.return_value = self.request_classes
        mock_commands.return_value = self.request_cmds
        mock_input.side_effect = ['testcommand arg1 arg2', EOFError]

        # This will loop twice. First loop should call _parse_command, making sure we pass the args
        # Second loop triggers exit
        _cli_loop(self.broker)
        mock_parse.assert_called_once_with(self.broker, self.request_classes, self.request_cmds, 'testcommand', ['arg1', 'arg2'])

    @patch('kafka.tools.protocol.__main__._get_request_classes')
    @patch('kafka.tools.protocol.__main__._get_request_commands')
    @patch('kafka.tools.protocol.__main__.input')
    @timeout_decorator.timeout(5)
    def test_cli_loop_exit(self, mock_input, mock_commands, mock_classes):
        mock_classes.return_value = self.request_classes
        mock_commands.return_value = self.request_cmds
        mock_input.return_value = 'exit'

        # This will should just exit the loop after 1 run
        _cli_loop(self.broker)

    @patch('kafka.tools.protocol.__main__._get_request_classes')
    @patch('kafka.tools.protocol.__main__._get_request_commands')
    @patch('kafka.tools.protocol.__main__.input')
    @timeout_decorator.timeout(5)
    def test_cli_loop_empty(self, mock_input, mock_commands, mock_classes):
        mock_classes.return_value = self.request_classes
        mock_commands.return_value = self.request_cmds
        mock_input.side_effect = ['', EOFError]

        # This should loop twice. The first loop is the test (empty input), the second triggers an exit of the loop
        _cli_loop(self.broker)

    @patch('kafka.tools.protocol.__main__._cli_loop')
    @patch('kafka.tools.protocol.__main__.set_up_arguments')
    @patch('kafka.tools.protocol.__main__.Broker')
    def test_main(self, mock_broker, mock_args, mock_cli_loop):
        mock_args.return_value = argparse.Namespace(broker='testhost', port=9328)
        mock_broker_inst = MagicMock()
        mock_broker.return_value = mock_broker_inst
        rv = main()
        assert rv == 0

        mock_args.assert_called_once_with()
        mock_broker.assert_called_once_with('testhost', port=9328)
        mock_broker_inst.connect.assert_called_once_with()
        mock_cli_loop.assert_called_once_with(mock_broker_inst)
        mock_broker_inst.close.assert_called_once_with()



import unittest
from mock import patch

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.leader_and_isr_v0 import _parse_argument, _process_arguments, LeaderAndIsrV0Request


class LeaderAndIsrV0RequestTests(unittest.TestCase):
    def setUp(self):
        self.values = {'controller_id': 5,
                       'controller_epoch': 42,
                       'partition_states': [],
                       'live_leaders': []}
        self.target = {'controller_id': 5,
                       'controller_epoch': 42,
                       'partition_states': [],
                       'live_leaders': []}

    def test_parse_argument_state(self):
        _parse_argument(self.values, 'topicname,3,53,2,64,6|7,23,8|9')

        self.target['partition_states'] = [{'topic': 'topicname',
                                            'partition': 3,
                                            'controller_epoch': 53,
                                            'leader': 2,
                                            'leader_epoch': 64,
                                            'isr': [6, 7],
                                            'zk_version': 23,
                                            'replicas': [8, 9]}]
        assert self.values == self.target

    def test_parse_argument_state_nonnumeric(self):
        self.assertRaises(ArgumentError, _parse_argument, self.values, 'topicname,foo,53,2,64,6|7,23,8|9')
        self.assertRaises(ArgumentError, _parse_argument, self.values, 'topicname,3,foo,2,64,6|7,23,8|9')
        self.assertRaises(ArgumentError, _parse_argument, self.values, 'topicname,3,53,foo,64,6|7,23,8|9')
        self.assertRaises(ArgumentError, _parse_argument, self.values, 'topicname,3,53,2,foo,6|7,23,8|9')
        self.assertRaises(ArgumentError, _parse_argument, self.values, 'topicname,3,53,2,64,foo|7,23,8|9')
        self.assertRaises(ArgumentError, _parse_argument, self.values, 'topicname,3,53,2,64,6|7,foo,8|9')
        self.assertRaises(ArgumentError, _parse_argument, self.values, 'topicname,3,53,2,64,6|7,23,foo|9')

    def test_parse_argument_live(self):
        _parse_argument(self.values, '3,examplehost,4398')

        self.target['live_leaders'] = [{'id': 3,
                                        'host': 'examplehost',
                                        'port': 4398}]
        assert self.values == self.target

    def test_parse_argument_live_nonnumeric(self):
        self.assertRaises(ArgumentError, _parse_argument, self.values, 'foo,examplehost,4398')
        self.assertRaises(ArgumentError, _parse_argument, self.values, '3,examplehost,foo')

    def test_parse_argument_notenough(self):
        self.assertRaises(ArgumentError, _parse_argument, self.values, '3,examplehost')

    def test_parse_argument_toomany(self):
        self.assertRaises(ArgumentError, _parse_argument, self.values, 'topicname,foo,53,2,64,6|7,23,8|9,foo')

    def test_process_arguments(self):
        val = _process_arguments("LeaderAndIsrV0", ['5', '42', 'topicname,3,53,2,64,6|7,23,8|9', '3,examplehost,4398'])

        self.target['partition_states'] = [{'topic': 'topicname',
                                            'partition': 3,
                                            'controller_epoch': 53,
                                            'leader': 2,
                                            'leader_epoch': 64,
                                            'isr': [6, 7],
                                            'zk_version': 23,
                                            'replicas': [8, 9]}]
        self.target['live_leaders'] = [{'id': 3,
                                        'host': 'examplehost',
                                        'port': 4398}]
        assert val == self.target

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, _process_arguments, "LeaderAndIsrV0", ['foo', '42', 'topicname,3,53,2,64,6|7,23,8|9', '3,examplehost,4398'])
        self.assertRaises(ArgumentError, _process_arguments, "LeaderAndIsrV0", ['5', 'foo', 'topicname,3,53,2,64,6|7,23,8|9', '3,examplehost,4398'])

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, _process_arguments, "LeaderAndIsrV0", ['foo', '42', 'topicname,3,53,2,64,6|7,23,8|9'])

    @patch('kafka.tools.protocol.requests.leader_and_isr_v0._process_arguments')
    def test_process_arguments_class(self, mock_process):
        mock_process.return_value = 'fake_return_value'
        val = LeaderAndIsrV0Request.process_arguments(['fake_arguments'])
        mock_process.assert_called_once_with("LeaderAndIsrV0", ['fake_arguments'])
        assert val == 'fake_return_value'



# -*- coding: utf-8 -*-
import sys, os, datetime

# Include source modules.
sys.path.insert(0, os.path.abspath("../"))

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.intersphinx']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
copyright = u'{year}, LinkedIn'.format(year=datetime.datetime.today().year)

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output ---------------------------------------------------

html_theme = 'default'
html_static_path = ['_static']

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'http://docs.python.org/': None}



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.list_groups_v0 import ListGroupsV0Request


class ListGroupsV0RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        assert ListGroupsV0Request.process_arguments([]) == {}

    def test_process_arguments_extra(self):
        self.assertRaises(ArgumentError, ListGroupsV0Request.process_arguments, ['foo'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from multiprocessing.pool import ThreadPool
from random import shuffle
from threading import RLock
import collections
import six
import time

from kafka.tools.configuration import ClientConfiguration
from kafka.tools.exceptions import ConnectionError, GroupError, TopicError, ConfigurationError
from kafka.tools.protocol.requests.describe_groups_v0 import DescribeGroupsV0Request
from kafka.tools.protocol.requests.group_coordinator_v0 import GroupCoordinatorV0Request
from kafka.tools.protocol.requests.list_groups_v0 import ListGroupsV0Request
from kafka.tools.protocol.requests.list_offset_v0 import ListOffsetV0Request
from kafka.tools.protocol.requests.offset_commit_v2 import OffsetCommitV2Request
from kafka.tools.protocol.requests.offset_fetch_v1 import OffsetFetchV1Request
from kafka.tools.protocol.requests.topic_metadata_v1 import TopicMetadataV1Request
from kafka.tools.models.broker import Broker
from kafka.tools.models.cluster import Cluster
from kafka.tools.models.group import Group
from kafka.tools.models.topic import Topic, TopicOffsets
from kafka.tools.utilities import synchronized, raise_if_error


class Client:
    # Special timestamps for offset requests
    OFFSET_EARLIEST = -2
    OFFSET_LATEST = -1

    def __init__(self, **kwargs):
        """
        Create a new Kafka client. There are two ways to instantiate the client:
            1) Provide a ClientConfiguration object with configuration=obj. This specifies all the configuration options
               for the client, or uses the defaults in ClientConfiguration. If the configuration keyword argument is
               present, all other arguments are ignored

            2) Provide keyword arguments that will be passed to create a new ClientConfiguration object. If the
               configuration keyword argument is not provided, this is the option that is chosen.

        The client will not be connected, either using the broker_list or the zkconnect option, until the connect()
        method is called

        Args:
            configuration (ClientConfiguration): a ClientConfiguration object to specify client settings
            kwargs: keyword arguments that are passed to create a new ClientConfiguration object

        Returns:
            Client: the Client object, ready for a connect call
        """
        if 'configuration' in kwargs:
            if not isinstance(kwargs['configuration'], ClientConfiguration):
                raise ConfigurationError("configuration object is not an instance of ClientConfiguration")
            self.configuration = kwargs['configuration']
        else:
            self.configuration = ClientConfiguration(**kwargs)

        self._controller_id = None
        self._last_full_metadata = 0.0
        self._last_group_list = 0.0
        self.cluster = Cluster()
        self._connected = False
        self._lock = RLock()

    @synchronized
    def connect(self):
        """
        Connect to the all cluster brokers and populate topic and partition information. If the client was created with
        a zkconnect string, first we connect to Zookeeper to bootstrap the broker and topic information from there,
        and then the client connects to all brokers in the cluster. Otherwise, connect to the bootstrap broker
        specified and fetch the broker and topic metadata for the cluster.

        Todo:
            * Currently assumes everything works. Need to check for failure to connect to
              ZK or bootstrap.
        """
        if self.configuration.zkconnect is not None:
            self.cluster = Cluster.create_from_zookeeper(zkconnect=self.configuration.zkconnect, fetch_topics=False)
            self._connected = True
        else:
            # Connect to bootstrap brokers until we succeed or exhaust the list
            try_brokers = list(self.configuration.broker_list)
            self._connected = False
            while (len(try_brokers) > 0) and not self._connected:
                self._connected = self._maybe_bootstrap_cluster(try_brokers.pop())

            if not self._connected:
                raise ConnectionError("Unable to bootstrap cluster information")

        # Connect to all brokers
        self._connect_all_brokers()

    @synchronized
    def close(self):
        """
        Close connections to all brokers. The configuration information is retained, so calling connect() again will
        reconnect to all brokers.
        """
        for broker_id in self.cluster.brokers:
            self.cluster.brokers[broker_id].close()
        self._connected = False

    @synchronized
    def list_topics(self, cache=True):
        """
        Get a list of all topics in the cluster

        Args:
            cache (boolean): If False, ignore cached metadata and always fetch from the cluster

        Raises:
            ConnectionError: If there is a failure to send the request to all brokers in the cluster
        """
        self._raise_if_not_connected()
        self._maybe_update_full_metadata(cache)
        return list(self.cluster.topics.keys())

    @synchronized
    def get_topic(self, topic_name, cache=True):
        """
        Get information on a topic in the cluster. If cache is True, used cached topic metadata
        if it has not expired. Otherwise, perform a metadata request to refresh information on
        the topic first.

        Args:
            topic_name (string): The name of the topic to return
            cache (boolean): If False, ignore cached metadata and always fetch from the cluster

        Returns:
            Topic: The Topic object holding the detail for the topic

        Raises:
            ConnectionError: If there is a failure to send the request to all brokers in the cluster
            TopicError: If the topic does not exist, or there is a problem getting metadata for it
        """
        self._raise_if_not_connected()
        try:
            topic = self.cluster.topics[topic_name]
            force_update = (not cache) or (not topic.updated_since(time.time() - self.configuration.metadata_refresh))
        except KeyError:
            force_update = True

        if force_update:
            # It doesn't matter what broker we fetch topic metadata from
            metadata = self._send_any_broker(TopicMetadataV1Request({'topics': [topic_name]}))
            raise_if_error(TopicError, metadata['topics'][0]['error'])

            self._update_from_metadata(metadata)

        return self.cluster.topics[topic_name]

    @synchronized
    def list_groups(self, cache=True):
        """
        Get a list of all topics in the cluster. This is done by sending a ListGroups request to
        every broker in the cluster and collating the responses.

        Args:
            cache (boolean): If False, ignore cached groups and always fetch from the cluster

        Returns:
            list (string): a list of valid group names in the cluster
            int: the number of brokers that failed to response
        """
        self._raise_if_not_connected()
        error_counter = self._maybe_update_groups_list(cache)
        return list(self.cluster.groups.keys()), error_counter

    @synchronized
    def get_group(self, group_name, cache=True):
        """
        Get information on a group in the cluster. If cache is True, used cached group information
        if it has not expired. Otherwise, send a request to refresh information on the group first.

        Args:
            group_name (string): The name of the group to return
            cache (boolean): If False, ignore cached information and always fetch from the cluster

        Returns:
            Group: The Group object holding the detail for the group

        Raises:
            ConnectionError: If there is a failure to send the request to all brokers in the cluster
            GroupError: If the group does not exist or there is a problem fetching information for it
        """
        self._raise_if_not_connected()
        try:
            group = self.cluster.groups[group_name]
            force_update = (not cache) or (not group.updated_since(time.time() - self.configuration.metadata_refresh))
        except KeyError:
            force_update = True

        if force_update:
            # Group detail must come from the coordinator broker
            group_info = self._send_group_aware_request(group_name, DescribeGroupsV0Request({'group_ids': [group_name]}))
            raise_if_error(GroupError, group_info['groups'][0]['error'])

            self._update_groups_from_describe(group_info)

        return self.cluster.groups[group_name]

    @synchronized
    def get_offsets_for_topic(self, topic_name, timestamp=OFFSET_LATEST):
        """
        Get the offsets for all the partitions in the specified topic. The offsets are requested at the specified
        timestamp, which defaults to the latest offsets available (defined as the offset of the next message to be
        produced).

        Args:
            topic_name (string or list): The name of the topic to fetch offsets for. For multiple topics, a list of
                topics can be provided
            timestamp (int): The timestamp (in millis) to fetch offsets at. If not specified, the default is the special
                timestamp OFFSET_LATEST, which requests the current end of the partitions (tail). You can also specify
                the special offset OFFSET_EARLIEST, which requests the oldest offset for each partition (head)

        Return:
            TopicOffsets: A TopicOffsets instance that contain offsets for all the partitions in the topic.

        Raises:
            ConnectionError: If there is a failure to send the request to a broker
            TopicError: If the topic does not exist or there is a problem getting information for it
            OffsetError: If there is a failure retrieving offsets for the specified topic or timestamp
            TypeError: If the timestamp is not an integer
        """
        return self.get_offsets_for_topics([topic_name], timestamp)[topic_name]

    @synchronized
    def get_offsets_for_topics(self, topic_list, timestamp=OFFSET_LATEST):
        """
        Get the offsets for all the partitions in the specified topics. The offsets are requested at the specified
        timestamp, which defaults to the latest offsets available (defined as the offset of the next message to be
        produced).

        Args:
            topic_list (list): a list of topic name strings to fetch offsets for.
            timestamp (int): The timestamp (in millis) to fetch offsets at. If not specified, the default is the special
                timestamp OFFSET_LATEST, which requests the current end of the partitions (tail). You can also specify
                the special offset OFFSET_EARLIEST, which requests the oldest offset for each partition (head)

        Return:
            dict (string -> TopicOffsets): A dictionary mapping topic names to TopicOffsets instances that contain
                offsets for all the partitions in the topic.

        Raises:
            ConnectionError: If there is a failure to send the request to a broker
            TopicError: If a topic does not exist or there is a problem getting information for it
            OffsetError: If there is a failure retrieving offsets for the specified topic or timestamp
            TypeError: If the timestamp is not an integer
        """
        self._raise_if_not_connected()
        if not isinstance(timestamp, six.integer_types):
            raise TypeError("timestamp must be a valid integer")

        # Get the topic information, making sure all the leadership info is current
        self._maybe_update_metadata_for_topics(topic_list)

        # Get a broker to topic-partition mapping
        broker_to_tp = self._map_topic_partitions_to_brokers(topic_list)

        request_values = {}
        for broker_id in broker_to_tp:
            request_values[broker_id] = {'replica_id': -1, 'topics': []}
            for topic_name in broker_to_tp[broker_id]:
                request_values[broker_id]['topics'].append({'topic': topic_name,
                                                            'partitions': [{'partition': i,
                                                                            'timestamp': timestamp,
                                                                            'max_num_offsets': 1} for i in broker_to_tp[broker_id][topic_name]]})

        return self._send_list_offsets_to_brokers(request_values)

    @synchronized
    def get_offsets_for_group(self, group_name, topic_list=None):
        """
        Get the latest offsets committed by the specified group. If a topic_name is specified, only the offsets for that
        topic are returned. Otherwise, offsets for all topics that the group is subscribed to are returned.

        Args:
            group_name (string): The name of the group to fetch offsets for
            topic_list (list): A list of string topic names to fetch offsets for. Defaults to None, which specifies all
                topics that are subscribed to by the group.

        Return:
            dict (string -> TopicOffsets): A dictionary mapping topic names to TopicOffsets instances that contain
                offsets for all the partitions in the topic. The dictionary will contain a single key if the topic_name
                provided is not None.

        Raises:
            ConnectionError: If there is a failure to send requests to brokers
            TopicError: If the topic does not exist or there is a problem getting information for it
            GroupError: If there is a failure to get information for the specified group
            OffsetError: If there is a failure retrieving offsets for the topic(s)
        """
        self._raise_if_not_connected()

        # Get the group we're fetching offsets for (potentially updating the group information)
        group = self.get_group(group_name)
        fetch_topics = self._get_topics_for_group(group, topic_list)

        # Get the topic information, making sure all the leadership info is current
        self._maybe_update_metadata_for_topics(fetch_topics)

        request_values = {'group_id': group_name,
                          'topics': [{'topic': topic, 'partitions': list(range(len(self.cluster.topics[topic].partitions)))} for topic in fetch_topics]}
        response = self._send_group_aware_request(group_name, OffsetFetchV1Request(request_values))

        rv = {}
        for topic in response['responses']:
            topic_name = topic['topic'].value()
            rv[topic_name] = TopicOffsets(self.cluster.topics[topic_name])
            rv[topic_name].set_offsets_from_fetch(topic['partition_responses'])

        return rv

    @synchronized
    def set_offsets_for_group(self, group_name, topic_offsets):
        """
        Given a group name and a list of topics and offsets, write the offsets as the latest for the group. This can only
        be done if the consumer is an old consumer (committing to Kafka), or if it is a new consumer in the "Empty"
        state (i.e. not running)

        Args:
            group_name (string): the name of the consumer group to set offsets for
            topic_offsets (list): a list of TopicOffsets objects which describe the topics and offsets to set

        Returns:
            dict (string -> list): a map of topic names to a list of error code responses for each partition. Error code
                0 indicates no error.

        Raises:
            ConnectionError: If there is a failure to send requests to brokers
            TypeError: if the topic_offsets argument is not properly formatted
            GroupError: If the group is not in the "Empty" state (if it is a new consumer), or if the group coordinator
                is unavailable
        """
        self._raise_if_not_connected()
        if isinstance(topic_offsets, six.string_types) or (not isinstance(topic_offsets, collections.Sequence)):
            raise TypeError("topic_offsets argument is not a list")

        # Get the group we're setting offsets for (potentially updating the group information)
        group = self.get_group(group_name)
        if group.state not in (None, 'Empty', 'Dead'):
            raise GroupError("The consumer group must be in the 'Empty' or 'Dead' state to set offsets, not '{0}'".format(group.state))

        # Get the topic information, making sure all the leadership info is current
        fetch_topics = [offsets.topic.name for offsets in topic_offsets]
        self._maybe_update_metadata_for_topics(fetch_topics)

        response = self._send_set_offset_request(group_name, topic_offsets)
        return self._parse_set_offset_response(response)

    # The following interfaces are for future implementation. The names are set, but the interfaces
    # are not yet
    #
    # def create_topic(self):
    #     raise NotImplementedError
    #
    # def delete_topic(self):
    #     raise NotImplementedError

    ##########################################################################
    # PRIVATE HELPER METHODS
    #
    # Everything below this point is methods that are not exposed, and are helpers for the
    # actual exposed interfaces.
    ##########################################################################

    def _raise_if_not_connected(self):
        if not self._connected:
            raise ConnectionError("The client is not yet connected")

    def _maybe_bootstrap_cluster(self, broker_port):
        """Attempt to bootstrap the cluster information using the given broker"""
        broker = Broker(broker_port[0], port=broker_port[1])

        try:
            broker.connect(sslcontext=self.configuration.ssl_context)
        except ConnectionError:
            # Just skip to the next bootstrap broker
            return False

        # Fetch topic metadata for all topics/brokers
        req = TopicMetadataV1Request({'topics': None})
        correlation_id, metadata = broker.send(req)
        broker.close()

        # Add brokers and topics to cluster
        self._controller_id = metadata['controller_id'].value()
        self._update_from_metadata(metadata)
        self._last_full_metadata = time.time()
        return True

    def _connect_all_brokers(self):
        for broker_id in self.cluster.brokers:
            self.cluster.brokers[broker_id].connect(self.configuration.ssl_context)

    def _send_to_broker(self, broker_id, request):
        """
        Given a broker ID and a request, send the request to that broker and return the response

        Args:
            broker_id (int): The ID of a broker in the cluster
            request (BaseRequest): A valid request object that inherits from BaseRequest

        Returns:
            BaseResponse: An instance of the response class associated with the request

        Raises:
            ConnectionError: If there is a failure to send the request to all brokers in the cluster
        """
        correlation_id, response = self.cluster.brokers[broker_id].send(request)
        return response

    def _send_any_broker(self, request):
        """
        Sends a request to any broker. This can be used for requests that do not require a
        specific broker to answer, such as metadata requests

        Args:
            request (BaseRequest): A valid request object that inherits from BaseRequest

        Returns:
            BaseResponse: An instance of the response class associated with the request

        Raises:
            ConnectionError: If there is a failure to send the request to all brokers in the cluster
        """
        broker_ids = list(self.cluster.brokers.keys())
        shuffle(broker_ids)
        while len(broker_ids) > 0:
            broker_id = broker_ids.pop()
            try:
                return self._send_to_broker(broker_id, request)
            except ConnectionError:
                # We're going to ignore failures unless we exhaust brokers
                pass

        # If we exhausted all the brokers, we have a serious problem
        raise ConnectionError("Failed to send request to any broker")

    def _send_some_brokers(self, requests, ignore_errors=True):
        """
        Sends a request to one or more brokers. The responses are returned mapped to the broker that
        they were retrieved from. This method uses a thread pool to parallelize sends.

        Args:
            request (int -> BaseRequest): A dictionary, where keys are integer broker IDs and the values are valid
                request objects that inherit from BaseRequest.

        Returns:
            dict (int -> BaseResponse): A map of broker IDs to response instances (inherited from
                BaseResponse). Failed requests are represented with a value of None
        """
        results = {}
        pool = ThreadPool(processes=self.configuration.broker_threads)
        for broker_id in requests:
            results[broker_id] = pool.apply_async(self._send_to_broker, (broker_id, requests[broker_id]))
        pool.close()
        pool.join()

        responses = {}
        for broker_id in results:
            try:
                responses[broker_id] = results[broker_id].get()
            except ConnectionError:
                if ignore_errors:
                    # Individual broker failures are OK, as we'll represent them with a None value
                    responses[broker_id] = None
                else:
                    raise
        return responses

    def _send_all_brokers(self, request):
        """
        Sends a request to all brokers. The responses are returned mapped to the broker that
        they were retrieved from

        Args:
            request (BaseRequest): A valid request object that inherits from BaseRequest

        Returns:
            dict (int -> BaseResponse): A map of broker IDs to response instances (inherited from
                BaseResponse). Failed requests are represented with a value of None
        """
        requests = {}
        for broker_id in self.cluster.brokers:
            requests[broker_id] = request
        return self._send_some_brokers(requests)

    def _send_group_aware_request(self, group_name, request):
        """
        Sends a request to the broker currently serving as the group coordinator for the specified
        group name. The request is not introspected for whether or not this is the correct broker,
        and the response is not checked as to whether or not there was an error.

        As a side effect of this call, the group object is created if it exists, and the coordinator
        attribute is set to the broker that is currently the coordinator for that group.

        Args:
            group_name (string): The name of the group to find the coordinator for
            request (BaseRequest): A request instance, inherited from BaseRequest, to send to the
                coordinator broker

        Returns:
            BaseResponse: The response instance, appropriate for the request, that is returned from
                the broker

        Raises:
            ConnectionError: If there is a failure to send the request to the coordinator broker,
                or a failure to retrieve the coordinator information
            GroupError: If an error is returned when fetching coordinator information
        """
        response = self._send_any_broker(GroupCoordinatorV0Request({'group_id': group_name}))
        raise_if_error(GroupError, response['error'])

        if group_name not in self.cluster.groups:
            self.cluster.add_group(Group(group_name))
        try:
            self.cluster.groups[group_name].coordinator = self.cluster.brokers[response['node_id'].value()]
        except KeyError:
            broker = Broker(response['host'].value(), id=response['node_id'].value(), port=response['port'].value())
            self.cluster.add_broker(broker)
            self.cluster.groups[group_name].coordinator = broker

        correlation_id, response = self.cluster.groups[group_name].coordinator.send(request)
        return response

    def _send_list_offsets_to_brokers(self, request_values):
        """
        Given a mapping of broker IDs to values for ListOffset requests, send the requests to all the brokers and
        collate the responses into a mapping of topic names to TopicOffsets instances

        Args:
            request_values (dict): a mapping of broker ID (int) to value dictionaries for ListOffsets requests

        Returns:
            dict (string -> TopicOffsets): A dictionary mapping topic names to TopicOffsets instances that contain
                offsets for all the partitions requested.

        Raises:
            ConnectionError: If there is a failure to send the request to a broker
            OffsetError: If there is a failure retrieving offsets
        """
        requests = {}
        for broker_id in request_values:
            requests[broker_id] = ListOffsetV0Request(request_values[broker_id])
        responses = self._send_some_brokers(requests, ignore_errors=False)

        rv = {}
        for broker_id in responses:
            response = responses[broker_id]
            for topic in response['responses']:
                topic_name = topic['topic'].value()
                if topic_name not in rv:
                    rv[topic_name] = TopicOffsets(self.cluster.topics[topic_name])
                rv[topic_name].set_offsets_from_list(topic['partition_responses'])

        return rv

    def _send_set_offset_request(self, group_name, topic_offsets):
        request = {'group_id': group_name,
                   'group_generation_id': -1,
                   'member_id': '',
                   'retention_time': -1,
                   'topics': []}
        for offset in topic_offsets:
            if not (isinstance(offset, TopicOffsets) and isinstance(offset.topic, Topic)):
                raise TypeError("TopicOffsets objects are not properly formed")

            request['topics'].append({'topic': offset.topic.name,
                                      'partitions': [{'partition': p_num,
                                                      'offset': p_offset,
                                                      'metadata': None} for p_num, p_offset in enumerate(offset.partitions)]})

        return self._send_group_aware_request(group_name, OffsetCommitV2Request(request))

    def _parse_set_offset_response(self, response):
        rv = {}
        for topic in response['responses']:
            topic_name = topic['topic'].value()
            rv[topic_name] = [-1] * len(topic['partition_responses'])
            for partition in topic['partition_responses']:
                rv[topic_name][partition['partition'].value()] = partition['error'].value()

        return rv

    def _update_brokers_from_metadata(self, metadata):
        """
        Given a Metadata response (either V0 or V1), update the broker information for this
        cluster. We don't delete brokers because we don't know if the brokers is gone temporarily
        (crashed or maintenance) or permanently.

        Args:
            metadata (MetadataV1Response): A metadata response to create or update brokers for
        """
        for b in metadata['brokers']:
            try:
                broker = self.cluster.brokers[b['node_id'].value()]
                if (broker.hostname != b['host'].value()) or (broker.port != b['port'].value()):
                    # if the hostname or port changes, close the existing connection
                    broker.close()
                    broker.hostname = b['host'].value()
                    broker.port = b['port'].value()
            except KeyError:
                broker = Broker(b['host'].value(), id=b['node_id'].value(), port=b['port'].value())
                self.cluster.add_broker(broker)
            broker.rack = b['rack'].value()

    def _maybe_delete_topics_not_in_metadata(self, metadata, delete):
        """
        If delete is True, check each topic in the cluster and delete it if it is not also in the metadata
        response provided. This should only be used when the metadata response has a full list of all topics
        in the cluster.

        Args:
            metadata (MetadataV1Response): a metadata response that contains the full list of topics in the
                cluster
            delete (boolean): If False, no action is taken
        """
        if not delete:
            return

        topic_list = metadata.topic_names()
        topics_for_deletion = []
        for topic_name in self.cluster.topics:
            if topic_name in topic_list:
                continue

            self.cluster.topics[topic_name].assure_has_partitions(0)
            topics_for_deletion.append(topic_name)

        for topic_name in topics_for_deletion:
            del self.cluster.topics[topic_name]

    def _update_topics_from_metadata(self, metadata, delete=False):
        """
        Given a Metadata response (either V0 or V1 will work), update the topic information
        for this cluster.

        Args:
            metadata (MetadataV1Response): A metadata response to create or update topics for
            delete (boolean): If True, delete topics from the cluster that are not present in the
                metadata response

        Raises:
            IndexError: If the brokers in the metadata object are not defined in the cluster
        """
        for t in metadata['topics']:
            if t['name'].value() not in self.cluster.topics:
                self.cluster.add_topic(Topic(t['name'].value(), len(t['partitions'])))
            topic = self.cluster.topics[t['name'].value()]
            topic._last_updated = time.time()

            topic.assure_has_partitions(len(t['partitions']))
            for p in t['partitions']:
                partition = topic.partitions[p['id'].value()]
                partition.leader = self.cluster.brokers[p['leader'].value()]
                for i, replica in enumerate(p['replicas']):
                    partition.add_or_update_replica(i, self.cluster.brokers[replica.value()])
                partition.delete_replicas(len(p['replicas']))

        self._maybe_delete_topics_not_in_metadata(metadata, delete)

    def _update_from_metadata(self, metadata, delete=False):
        """
        Given a metadata response, update both the brokers and topics from it. If specified, delete the topics
        that are not present in the provided metadata

        Args:
            metadata (MetadataV1Response): A metadata response to create or update brokers and topics for
            delete (boolean): If True, delete topics from the cluster that are not present in the metadata response
        """
        self._update_brokers_from_metadata(metadata)
        self._update_topics_from_metadata(metadata, delete=delete)

    def _maybe_update_metadata_for_topics(self, topics, cache=True):
        """
        Fetch metadata for the topics specified from the cluster if cache is False or if the cached metadata has expired.
        The cluster brokers and topics are updated with the metadata information.

        Args:
            cache (boolean): Ignore the metadata expiration and fetch from the cluster anyway

        Raises:
            ConnectionError: If there is a failure to send the request to all brokers in the cluster
        """
        force_update = not cache
        for topic in topics:
            try:
                if not self.cluster.topics[topic].updated_since(time.time() - self.configuration.metadata_refresh):
                    force_update = True
            except KeyError:
                force_update = True

        if force_update:
            self._update_from_metadata(self._send_any_broker(TopicMetadataV1Request({'topics': topics})), delete=False)

    def _maybe_update_full_metadata(self, cache=True):
        """
        Fetch all topic metadata from the cluster if cache is False or if the cached metadata has expired. The cluster
        brokers and topics are updated with the metadata information, potentially deleting ones that no longer exist

        Args:
            cache (boolean): Ignore the metadata expiration and fetch from the cluster anyway

        Raises:
            ConnectionError: If there is a failure to send the request to all brokers in the cluster
        """
        if (not cache) or (self._last_full_metadata < (time.time() - self.configuration.metadata_refresh)):
            self._update_from_metadata(self._send_any_broker(TopicMetadataV1Request({'topics': None})), delete=True)
            self._last_full_metadata = time.time()

    def _add_or_update_group(self, group_info, coordinator):
        """
        Given group information from a ListGroups response, assure that the group exists in the cluster as specified

        Args:
            group_info (dict): A group from a ListGroups response, which contains group_id and protocol_type keys
            coordinator (int): The ID of the group coordinator broker
        """
        group_name = group_info['group_id'].value()
        try:
            group = self.cluster.groups[group_name]
        except KeyError:
            group = Group(group_name)
            self.cluster.add_group(group)
        group.coordinator = self.cluster.brokers[coordinator]
        group.protocol_type = group_info['protocol_type'].value()

    def _update_groups_from_lists(self, responses):
        """
        Given a list of ListGroups responses, make sure that all the groups are in the cluster correctly

        Args:
            responses (dict): a mapping of broker IDs to ListGroupsV0Response instances from each broker

        Returns:
            int: a count of the number of responses that were not present or in error
        """
        error_counter = 0
        for broker_id, response in responses.items():
            if (response is None) or (response['error'].value() != 0):
                error_counter += 1
                continue

            for group_info in response['groups']:
                self._add_or_update_group(group_info, broker_id)

        return error_counter

    def _maybe_update_groups_list(self, cache=True):
        """
        Fetch lists of groups from all brokers if cache is False or if the cached group list has expired.

        Args:
            cache (boolean): Ignore the metadata expiration and fetch from the cluster anyway

        Returns:
            int: a count of the number of brokers that returned no response or error responses
        """
        if (not cache) or (self._last_group_list < (time.time() - self.configuration.metadata_refresh)):
            error_counter = self._update_groups_from_lists(self._send_all_brokers(ListGroupsV0Request({})))
            self._last_group_list = time.time()
            return error_counter
        return 0

    def _update_groups_from_describe(self, response):
        """
        Given a DescribeGroupsV0 response, update the group information for all groups in the response
        in this cluster. This does not delete any groups not in the response, as we do not fetch all
        describe groups data like that at this time

        Args:
            response (DescribeGroupsV0Response): A response to create or update groups for
        """
        for g in response['groups']:
            if g['group_id'].value() not in self.cluster.groups:
                self.cluster.add_group(Group(g['group_id'].value()))
            group = self.cluster.groups[g['group_id'].value()]
            group.state = g['state'].value()
            group.protocol_type = g['protocol_type'].value()
            group.protocol = g['protocol'].value()
            group.clear_members()
            for m in g['members']:
                group.add_member(m['member_id'].value(),
                                 client_id=m['client_id'].value(),
                                 client_host=m['client_host'].value(),
                                 metadata=m['member_metadata'].value(),
                                 assignment=m['member_assignment'].value())
            group._last_updated = time.time()

    def _map_topic_partitions_to_brokers(self, topic_list):
        """
        Given a list of topics, map the topic-partitions to the leader brokers

        Args:
            topic_list (list): a list of the topic name strings to map

        Returns:
            dict: a multi-dimensional dictionary where the keys are broker IDs and the values are a dictionary where the
                keys are topic names and the values are an array of partition IDs

        Raises:
            TopicError: if any of the topics do not exist in the cluster
        """
        broker_to_tp = {}
        for topic_name in topic_list:
            try:
                topic = self.cluster.topics[topic_name]
            except KeyError:
                raise TopicError("Topic {0} does not exist in the cluster".format(topic_name))

            for i, partition in enumerate(topic.partitions):
                broker_to_tp.setdefault(partition.leader.id, {}).setdefault(topic_name, []).append(i)

        return broker_to_tp

    def _get_topics_for_group(self, group, topic_list):
        """
        Given a group and a topic_list, return a list of topics that is either the topic list provided or the list of
        all topics consumed by the group

        Args:
            group (Group): the group for which the topic list is being created
            topic_list (list): A list of string topic names to fetch offsets for. Defaults to None, which specifies all
                topics that are subscribed to by the group.

        Returns:
             list: a list of topics to be fetched for the group

        Raises:
            GroupError: if the topic list is empty
        """
        # We'll be nice. If the topic_list is a string, convert it to a list
        if isinstance(topic_list, six.string_types):
            topic_list = [topic_list]

        # Create a list of topics to fetch
        fetch_topics = topic_list or group.subscribed_topics()
        if len(fetch_topics) == 0:
            raise GroupError("No topic specified is consumed by the group")
        return fetch_topics



import unittest

import kafka.tools.exceptions


class ExceptionTests(unittest.TestCase):
    def test_exception_assigner_plain(self):
        e = kafka.tools.exceptions.AssignerException()
        assert str(e) == "Unknown Assigner Exception"
        assert isinstance(e, Exception)

    def test_exception_assigner_custom(self):
        e = kafka.tools.exceptions.AssignerException("custom text")
        assert str(e) == "custom text"

    def test_exception_notfound(self):
        e = kafka.tools.exceptions.ReplicaNotFoundException()
        assert isinstance(e, kafka.tools.exceptions.AssignerException)

    def test_exception_notenough(self):
        e = kafka.tools.exceptions.NotEnoughReplicasException()
        assert isinstance(e, kafka.tools.exceptions.AssignerException)

    def test_exception_config(self):
        e = kafka.tools.exceptions.ConfigurationException()
        assert isinstance(e, kafka.tools.exceptions.AssignerException)

    def test_exception_zookeeper(self):
        e = kafka.tools.exceptions.ZookeeperException()
        assert isinstance(e, kafka.tools.exceptions.AssignerException)

    def test_exception_consistency(self):
        e = kafka.tools.exceptions.ClusterConsistencyException()
        assert isinstance(e, kafka.tools.exceptions.AssignerException)

    def test_exception_programming(self):
        e = kafka.tools.exceptions.ProgrammingException()
        assert isinstance(e, kafka.tools.exceptions.AssignerException)

    def test_exception_reassignment_failed(self):
        e = kafka.tools.exceptions.ReassignmentFailedException()
        assert isinstance(e, kafka.tools.exceptions.AssignerException)

    def test_exception_balance(self):
        e = kafka.tools.exceptions.BalanceException()
        assert isinstance(e, kafka.tools.exceptions.AssignerException)

    def test_exception_unknown_broker(self):
        e = kafka.tools.exceptions.UnknownBrokerException()
        assert isinstance(e, kafka.tools.exceptions.AssignerException)



import unittest

from kafka.tools.models.group import Group, GroupMember


class GroupTests(unittest.TestCase):
    def test_group_create(self):
        group = Group('testgroup')
        assert group.name == 'testgroup'

    def test_updated_since(self):
        group = Group('testgroup')
        group._last_updated = 100
        assert group.updated_since(99)

    def test_add_member(self):
        group = Group('testgroup')
        group.add_member('membername', client_id='clientid', client_host='host1', metadata=b'\x00\x32', assignment=b'\x01\x34')
        assert len(group.members) == 1
        assert group.members[0].name == 'membername'
        assert group.members[0].client_id == 'clientid'
        assert group.members[0].client_host == 'host1'
        assert group.members[0].metadata == b'\x00\x32'
        assert group.members[0].assignment_data == b'\x01\x34'

    def test_clear_members(self):
        group = Group('testgroup')
        group.members = 'baddata'
        group.clear_members()
        assert group.members == []

    def test_set_assignment(self):
        group = Group('testgroup')
        group.protocol_type = 'consumer'
        member = GroupMember('membername', assignment=b'\x00\x00\x00\x00\x00\x01\x00\x06topic1\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x02\x65\xbd')
        member.group = group

        member.set_assignment()
        assert member.assignment_version == 0
        assert member.assignment_data == b'\x00\x00\x00\x00\x00\x01\x00\x06topic1\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x02\x65\xbd'
        assert member.user_data == b'\x65\xbd'
        assert member.topics == {'topic1': [0]}

    def test_subscribed_topics(self):
        group = Group('testgroup')
        group.protocol_type = 'consumer'
        member = GroupMember('member1')
        member.group = group
        member.topics = {'topic1': [0]}
        group.members.append(member)
        member = GroupMember('member2')
        member.group = group
        member.topics = {'topic2': [0], 'topic1': [1]}
        group.members.append(member)

        topics = group.subscribed_topics()
        print(topics)
        assert set(['topic1', 'topic2']) == set(topics)



import unittest

from kafka.tools.exceptions import ReplicaNotFoundException, ClusterConsistencyException
from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.models.cluster import Cluster


class PartitionOperationTests(unittest.TestCase):
    def setUp(self):
        self.cluster = Cluster()
        self.cluster.add_broker(Broker("brokerhost1.example.com", id=1))
        self.cluster.add_broker(Broker("brokerhost2.example.com", id=2))
        self.cluster.add_broker(Broker("brokerhost3.example.com", id=3))
        self.cluster.add_topic(Topic("testTopic1", 2))
        self.cluster.add_topic(Topic("testTopic2", 2))

    def add_topics(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition.add_replica(self.cluster.brokers[2], 1)
        partition = self.cluster.topics['testTopic1'].partitions[1]
        partition.add_replica(self.cluster.brokers[2], 0)
        partition.add_replica(self.cluster.brokers[1], 1)
        partition = self.cluster.topics['testTopic2'].partitions[0]
        partition.add_replica(self.cluster.brokers[2], 0)
        partition.add_replica(self.cluster.brokers[1], 1)
        partition = self.cluster.topics['testTopic2'].partitions[1]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition.add_replica(self.cluster.brokers[2], 1)

    def test_partition_add_broker_partition(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition._add_broker_partition(0, self.cluster.brokers[1])
        assert self.cluster.brokers[1].partitions[0] == [partition]

    def test_partition_add_broker_partition_two(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition._add_broker_partition(0, self.cluster.brokers[1])

        partition2 = self.cluster.topics['testTopic2'].partitions[1]
        partition2._add_broker_partition(0, self.cluster.brokers[1])
        assert self.cluster.brokers[1].partitions[0] == [partition, partition2]

    def test_partition_add_replica(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        assert self.cluster.brokers[1].partitions[0] == [partition]
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [self.cluster.brokers[1]]

    def test_partition_add_replica_two(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition.add_replica(self.cluster.brokers[2], 1)
        assert self.cluster.brokers[1].partitions[0] == [partition]
        assert self.cluster.brokers[2].partitions[1] == [partition]
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [self.cluster.brokers[1], self.cluster.brokers[2]]

    def test_partition_remove_broker_partition(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition._remove_broker_partition(self.cluster.brokers[1])
        assert self.cluster.brokers[1].partitions[0] == []

    def test_partition_remove_broker_partition_two(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition2 = self.cluster.topics['testTopic2'].partitions[1]
        partition2.add_replica(self.cluster.brokers[1], 0)
        partition._remove_broker_partition(self.cluster.brokers[1])
        assert self.cluster.brokers[1].partitions[0] == [partition2]

    def test_partition_remove_replica(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition.remove_replica(self.cluster.brokers[1])
        assert self.cluster.brokers[1].partitions[0] == []
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == []

    def test_partition_remove_replica_single(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition.add_replica(self.cluster.brokers[2], 1)
        partition.remove_replica(self.cluster.brokers[1])
        assert self.cluster.brokers[1].partitions[0] == []
        assert self.cluster.brokers[2].partitions[1] == [partition]
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [self.cluster.brokers[2]]

    def test_partition_remove_replica_nonexistent(self):
        self.assertRaises(ReplicaNotFoundException, self.cluster.topics['testTopic1'].partitions[0].remove_replica, self.cluster.brokers[1])

    def test_partition_swap_replicas(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition.swap_replicas(self.cluster.brokers[1], self.cluster.brokers[2])
        assert self.cluster.brokers[2].partitions[0] == [partition]
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [self.cluster.brokers[2]]

    def test_partition_swap_replicas_nonexistent(self):
        self.assertRaises(ReplicaNotFoundException,
                          self.cluster.topics['testTopic1'].partitions[0].swap_replicas,
                          self.cluster.brokers[1],
                          self.cluster.brokers[2])

    def test_partition_swap_replica_positions(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition.add_replica(self.cluster.brokers[2], 1)
        partition.swap_replica_positions(self.cluster.brokers[1], self.cluster.brokers[2])
        assert self.cluster.brokers[2].partitions[0] == [partition]
        assert self.cluster.brokers[1].partitions[1] == [partition]
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [self.cluster.brokers[2], self.cluster.brokers[1]]

    def test_partition_swap_replica_positions_nonexistent(self):
        partition = self.cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition.add_replica(self.cluster.brokers[2], 1)
        self.assertRaises(ReplicaNotFoundException, partition.swap_replica_positions, self.cluster.brokers[1], self.cluster.brokers[3])

    def test_cluster_clone(self):
        # Should have a consistent cluster state
        self.add_topics()

        newcluster = self.cluster.clone()
        assert self.cluster is not newcluster
        for bid in newcluster.brokers:
            assert newcluster.brokers[bid] == self.cluster.brokers[bid]
        for tname in newcluster.topics:
            assert newcluster.topics[tname] == self.cluster.topics[tname]
        for partition in newcluster.partitions([]):
            assert partition == self.cluster.topics[partition.topic.name].partitions[partition.num]
            assert partition.replicas == self.cluster.topics[partition.topic.name].partitions[partition.num].replicas

    def test_cluster_changed_partitions(self):
        self.add_topics()
        newcluster = self.cluster.clone()
        newcluster.topics['testTopic1'].partitions[0].replicas.reverse()
        difference = self.cluster.changed_partitions(newcluster)
        assert difference == [newcluster.topics['testTopic1'].partitions[0]]

    def test_cluster_changed_partitions_inconsistent(self):
        self.add_topics()
        badcluster = Cluster()
        self.assertRaises(ClusterConsistencyException, badcluster.changed_partitions, self.cluster)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.types import BaseType
from kafka.tools.protocol.types.integers import Int8


class Boolean(BaseType):
    _type = "boolean"

    def _validate(self):
        if self._value:
            self._value = True
            self._underlying_value = Int8(1)
        else:
            self._value = False
            self._underlying_value = Int8(0)

    @classmethod
    def decode(cls, byte_array, schema=None):
        underlying_value, remaining_bytes = Int8.decode(byte_array)
        return cls(underlying_value.value() == 1, schema=cls._type), remaining_bytes

    def encode(self):
        return self._underlying_value.encode()

    def __bool__(self):
        return self._value

    def __nonzero__(self):
        return self.__bool__()

    def __str__(self):
        return '{0} (boolean)'.format(self._value)

    def __repr__(self):
        return '<Boolean {0}>'.format(self._value)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import six
import ssl

from kafka.tools.exceptions import ConfigurationError


def eval_boolean(value):
    """Attempt to evaluate the argument as a boolean"""
    if isinstance(value, bool):
        return value
    elif isinstance(value, six.integer_types):
        return value != 0
    else:
        return value.lower() in ['true', 'yes', 'on']


def check_file_access(filename):
    if not os.access(filename, os.R_OK):
        raise ConfigurationError("filename specified ({0}) is not accessible for reading".format(filename))


class ClientConfiguration(object):
    #######################
    # CONNECTION PROPERTIES
    #
    # broker_list and zkconnect are the two possible ways of specifying the Kafka cluster to connect to. One of these
    # options must be provided, and only one can be provided. By default, the broker_list is set to point to localhost

    @property
    def broker_list(self):
        """The broker list to use for bootstrapping the client

        This broker list is only used for the initial client connect. The client will connect to one of the brokers
        specified and fetch the cluster information, including a list of brokers.

        The format of the broker list is a comma-separated list of hostname:port
            hostname:port[,hostname:port...]
        """
        return getattr(self, '_broker_list', 'localhost:9092')

    @broker_list.setter
    def broker_list(self, value):
        if not isinstance(value, six.string_types):
            raise TypeError("broker_list must be a string")

        # We're not going to validate much here - if the user does the wrong thing, they'll get an error on connect
        self._broker_list = []
        hostports = value.split(',')
        for hostport in hostports:
            (host, port) = hostport.rsplit(':', 1)
            self._broker_list.append((host, int(port)))

    @property
    def zkconnect(self):
        """The zookeeper connection string for the Kafka cluster

        This is the Zookeeper connection string that points to the Kafka cluster metadata. It is the same stirng that
        is used when configuring the Kafka brokers. The format is:
            host:port[,host:port...][/chroot/path]
        """
        return getattr(self, '_zkconnect', None)

    @zkconnect.setter
    def zkconnect(self, value):
        # We're not going to validate this past being a string. It's too much of a pain, and Kazoo will handle it
        if not isinstance(value, six.string_types):
            raise TypeError("zkconnect must be a string")
        self._zkconnect = value

    ################
    # SSL PROPERTIES
    #
    # All of these properties are using for setting up TLS connections to the Kafka brokers. The defaults provided are
    # reasonable for a secure connection, except that enable_tls is disabled by default.

    @property
    def enable_tls(self):
        """Enable TLS for Kafka broker connections"""
        return getattr(self, '_enable_tls', False)

    @enable_tls.setter
    def enable_tls(self, value):
        self._enable_tls = eval_boolean(value)

    @property
    def tls_verify_certificates(self):
        """Define whether or not to verify the server host certificate is valid and trusted when TLS is enabled"""
        return getattr(self, '_tls_verify_certificates', True)

    @tls_verify_certificates.setter
    def tls_verify_certificates(self, value):
        self._tls_verify_certificates = eval_boolean(value)

    @property
    def tls_verify_hostnames(self):
        """Define whether or not to verify the server hostname matches the host certificate provided"""
        return getattr(self, '_tls_verify_hostnames', True)

    @tls_verify_hostnames.setter
    def tls_verify_hostnames(self, value):
        self._tls_verify_hostnames = eval_boolean(value)

    @property
    def tls_root_certificates(self):
        """Path to the trusted root certificates. If not provided, the system default will be used"""
        return getattr(self, '_tls_root_certificates', None)

    @tls_root_certificates.setter
    def tls_root_certificates(self, value):
        check_file_access(value)
        self._tls_root_certificates = value

    @property
    def tls_client_certificate(self):
        """Path to the client certificate, optionally including a key. If not provided, no client certificate is used"""
        return getattr(self, '_tls_client_certificate', None)

    @tls_client_certificate.setter
    def tls_client_certificate(self, value):
        check_file_access(value)
        self._tls_client_certificate = value

    @property
    def tls_client_keyfile(self):
        """Path to the client certificate key file, if separate from the client certificate file."""
        return getattr(self, '_tls_client_keyfile', None)

    @tls_client_keyfile.setter
    def tls_client_keyfile(self, value):
        check_file_access(value)
        self._tls_client_keyfile = value

    @property
    def tls_client_key_password_callback(self):
        """A function that will be called to get the keyfile password.
        This is a function that will be called to get the password that protects the keyfile specified. This must be a
        Python callable that takes no arguments. It must return a string, byte, or bytearray

        If not specified, the keyfile is assumed to be unprotected
        """
        return getattr(self, '_tls_client_key_password_callback', None)

    @tls_client_key_password_callback.setter
    def tls_client_key_password_callback(self, value):
        if not callable(value):
            raise TypeError("tls_client_key_password_callback is not callable".format(value))
        self._tls_client_key_password_callback = value

    #######################
    # KAFKA CLIENT SETTINGS
    #
    # The rest of these configurations are used for controlling the behavior of the client

    @property
    def metadata_refresh(self):
        """How long topic and group metadata can be cached"""
        return getattr(self, '_metadata_refresh', 60000)

    @metadata_refresh.setter
    def metadata_refresh(self, value):
        raise_if_not_positive_integer("metadata_refresh", value)
        self._metadata_refresh = value

    @property
    def broker_threads(self):
        """How many threads to use in a pool for broker connections"""
        return getattr(self, '_broker_threads', 20)

    @broker_threads.setter
    def broker_threads(self, value):
        raise_if_not_positive_integer("broker_threads", value)
        self._broker_threads = value

    def _set_attributes(self, **kwargs):
        for key in kwargs:
            if not hasattr(self, key):
                raise ConfigurationError("Invalid configuration specified: {0}".format(key))
            setattr(self, key, kwargs[key])

    def __init__(self, **kwargs):
        """
        Create a configuration object, setting any provided options. Either broker_list or zkconnect (but not both)
        must be provided

        Raises:
            ConfigurationError: unless exactly one of broker_list or zkconnect is provided, or if any invalid option
                is specified.
        """
        if ('zkconnect' in kwargs) and ('broker_list' in kwargs):
            raise ConfigurationError("Only one of zkconnect and broker_list may be provided")
        self._set_attributes(**kwargs)

        # Create the SSL context if we are going to enable TLS
        self.ssl_context = self._create_ssl_context() if self.enable_tls else None

    def _verify_ssl_configuration(self):
        if self.tls_verify_hostnames and (not self.tls_verify_certificates):
            raise ConfigurationError("tls_verify_hostnames may not be specified if tls_verify_certificates is False")

    def _create_ssl_context(self):
        self._verify_ssl_configuration()

        try:
            context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            context.verify_mode = ssl.CERT_REQUIRED if self.tls_verify_certificates else ssl.CERT_NONE
            context.check_hostname = self.tls_verify_hostnames

            if self.tls_root_certificates is None:
                context.load_default_certs(purpose=ssl.Purpose.CLIENT_AUTH)
            else:
                context.load_verify_locations(cafile=self.tls_root_certificates)

            if self.tls_client_certificate is not None:
                context.load_cert_chain(self.tls_client_certificate,
                                        keyfile=self.tls_client_keyfile,
                                        password=self.tls_client_key_password_callback)
        except ssl.SSLError as e:
            raise ConfigurationError("Unable to configure SSL Context: {0}".format(e))

        return context


def raise_if_not_positive_integer(attr_name, value):
    if not (isinstance(value, six.integer_types) and (value > 0)):
        raise TypeError("{0} must be a positive integer".format(attr_name))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import binascii

from kafka.tools.protocol.requests import BaseRequest, ArgumentError
from kafka.tools.protocol.responses.join_group_v0 import JoinGroupV0Response


def _parse_group_protocol(protocol):
    pparts = protocol.split(',')
    if len(pparts) != 2:
        raise ArgumentError("Expected group_protocol_name,group_protocol_metadata")
    if pparts[1] == '':
        return {'protocol_name': pparts[0], 'protocol_metadata': None}
    else:
        try:
            return {'protocol_name': pparts[0], 'protocol_metadata': binascii.unhexlify(pparts[1])}
        except:
            raise ArgumentError("group_protocol_metadata must be empty or a hex string")


class JoinGroupV0Request(BaseRequest):
    api_key = 11
    api_version = 0
    cmd = "JoinGroup"
    response = JoinGroupV0Response

    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} group_id session_timeout member_id protocol_type ".format(cmd, api_version) +
                   "(group_protocol_name,group_protocol_metadata ...)\n" +
                   "Description: Join or create a consumer group\n")

    schema = [
        {'name': 'group_id', 'type': 'string'},
        {'name': 'session_timeout', 'type': 'int32'},
        {'name': 'member_id', 'type': 'string'},
        {'name': 'protocol_type', 'type': 'string'},
        {'name': 'group_protocols',
         'type': 'array',
         'item_type': [
             {'name': 'protocol_name', 'type': 'string'},
             {'name': 'protocol_metadata', 'type': 'bytes'},
         ]},
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        if len(cmd_args) < 5:
            raise ArgumentError("JoinGroupV0 requires at least 5 arguments")

        try:
            values = {'group_id': cmd_args[0],
                      'session_timeout': int(cmd_args[1]),
                      'member_id': cmd_args[2],
                      'protocol_type': cmd_args[3],
                      'group_protocols': []}
        except ValueError:
            raise ArgumentError("The session_timeout must be an integer")

        for protocol in cmd_args[4:]:
            values['group_protocols'].append(_parse_group_protocol(protocol))

        return values



import six
import unittest

from kafka.tools.protocol.types.integers import Int8, _get_other_value


class BaseIntegerTests(unittest.TestCase):
    def test_get_other_value(self):
        val = _get_other_value(Int8(42))
        assert isinstance(val, six.integer_types)
        assert val == 42

    def test_get_other_value_int(self):
        val = _get_other_value(42)
        assert isinstance(val, six.integer_types)
        assert val == 42

    def test_get_bad_type(self):
        self.assertRaises(TypeError, _get_other_value, 'foo')

    def test_eq(self):
        val = Int8(123)
        val2 = Int8(123)
        assert val == val2
        assert val == 123
        assert 123 == val

    def test_ne(self):
        val = Int8(123)
        val2 = Int8(91)
        assert val != val2
        assert val != 91
        assert 91 != val

    def test_lt(self):
        val = Int8(91)
        val2 = Int8(123)
        assert val < val2
        assert val < 123
        assert 91 < val2

    def test_le(self):
        val = Int8(91)
        val2 = Int8(123)
        val3 = Int8(91)
        assert val <= val2
        assert val <= val3
        assert val <= 123
        assert val2 <= 123
        assert 91 <= val2
        assert 91 <= val

    def test_gt(self):
        val = Int8(91)
        val2 = Int8(123)
        assert val2 > val
        assert val2 > 91
        assert 123 > val

    def test_ge(self):
        val = Int8(91)
        val2 = Int8(123)
        val3 = Int8(91)
        assert val2 >= val
        assert val3 >= val
        assert val2 >= 91
        assert val >= 91
        assert 123 >= val
        assert 91 >= val

    def test_bool(self):
        val = Int8(0)
        val2 = Int8(1)
        assert not val
        assert val2

    def test_str(self):
        val = Int8(123)
        assert str(val) == "123 (int8)"
        assert repr(val) == "<Int8 123>"



import unittest

from kafka.tools.protocol.types.integers import Int16


class Int16Tests(unittest.TestCase):
    def test_create(self):
        val = Int16(23865)
        assert val.value() == 23865

    def test_create_null(self):
        self.assertRaises(TypeError, Int16, None)

    def test_encode_bad_type(self):
        self.assertRaises(TypeError, Int16, 'abc')

    def test_encode_bad_range(self):
        self.assertRaises(ValueError, Int16, 57463)

    def test_decode(self):
        (val, rest) = Int16.decode(b']9')
        assert isinstance(val, Int16)
        assert val.value() == 23865
        assert rest == b''

    def test_decode_remainder(self):
        (val, rest) = Int16.decode(b']9abc')
        assert val.value() == 23865
        assert rest == b'abc'

    def test_decode_underflow(self):
        self.assertRaises(ValueError, Int16.decode, b'')

    def test_encode(self):
        val = Int16(23865)
        assert val.encode() == b']9'



from kafka.tools.protocol.responses.list_groups_v0 import ListGroupsV0Response
from kafka.tools.protocol.responses.list_offset_v0 import ListOffsetV0Response
from kafka.tools.protocol.responses.offset_commit_v2 import OffsetCommitV2Response
from kafka.tools.protocol.responses.offset_fetch_v1 import OffsetFetchV1Response
from kafka.tools.protocol.responses.describe_groups_v0 import DescribeGroupsV0Response
from kafka.tools.protocol.responses.metadata_v1 import MetadataV1Response
from kafka.tools.protocol.responses.metadata_v0 import MetadataV0Response
from kafka.tools.protocol.responses.group_coordinator_v0 import GroupCoordinatorV0Response


def describe_groups():
    ma_1 = b'\x00\x00\x00\x00\x00\x01\x00\x06topic1\x00\x00\x00\x01\x00\x00\x00\x00\xff\xff\xff\xff'
    ma_2 = b'\x00\x00\x00\x00\x00\x01\x00\x06topic1\x00\x00\x00\x01\x00\x00\x00\x01\xff\xff\xff\xff'
    return DescribeGroupsV0Response.from_dict({'groups': [{'group_id': 'testgroup',
                                                           'error': 0,
                                                           'state': 'Stable',
                                                           'protocol': 'roundrobin',
                                                           'protocol_type': 'consumer',
                                                           'members': [{'member_id': 'testmember1',
                                                                        'client_id': 'testclientid1',
                                                                        'client_host': 'host.example.com',
                                                                        'member_metadata': b'\x90\x83\x24\xbc',
                                                                        'member_assignment': ma_1},
                                                                       {'member_id': 'testmember2',
                                                                        'client_id': 'testclientid2',
                                                                        'client_host': 'otherhost.example.com',
                                                                        'member_metadata': b'\x89\x34\x78\xad',
                                                                        'member_assignment': ma_2}
                                                                       ]
                                                           }]
                                               })


def describe_groups_error():
    return DescribeGroupsV0Response.from_dict({'groups': [{'group_id': 'testgroup',
                                                           'error': 16,
                                                           'state': None,
                                                           'protocol': None,
                                                           'protocol_type': None,
                                                           'members': None}]})


def topic_metadata():
    return MetadataV1Response.from_dict({'brokers': [{'node_id': 1,
                                                      'host': 'host1.example.com',
                                                      'port': 8031,
                                                      'rack': 'rack1'},
                                                     {'node_id': 101,
                                                      'host': 'host2.example.com',
                                                      'port': 8032,
                                                      'rack': 'rack2'}],
                                         'controller_id': 1,
                                         'topics': [{'error': 0,
                                                     'name': 'topic1',
                                                     'internal': False,
                                                     'partitions': [{'error': 0,
                                                                     'id': 0,
                                                                     'leader': 1,
                                                                     'replicas': [101, 1],
                                                                     'isrs': [101, 1]},
                                                                    {'error': 0,
                                                                     'id': 1,
                                                                     'leader': 101,
                                                                     'replicas': [101, 1],
                                                                     'isrs': [1, 101]}]}]})


def topic_metadata_v0():
    return MetadataV0Response.from_dict({'brokers': [{'node_id': 1,
                                                      'host': 'host1.example.com',
                                                      'port': 8031,
                                                      'rack': 'rack1'},
                                                     {'node_id': 101,
                                                      'host': 'host2.example.com',
                                                      'port': 8032,
                                                      'rack': 'rack2'}],
                                         'controller_id': 1,
                                         'topics': [{'error': 0,
                                                     'name': 'topic1',
                                                     'internal': False,
                                                     'partitions': [{'error': 0,
                                                                     'id': 0,
                                                                     'leader': 1,
                                                                     'replicas': [101, 1],
                                                                     'isrs': [101, 1]},
                                                                    {'error': 0,
                                                                     'id': 1,
                                                                     'leader': 101,
                                                                     'replicas': [101, 1],
                                                                     'isrs': [1, 101]}]}]})


def topic_metadata_error():
    return MetadataV1Response.from_dict({'brokers': [{'node_id': 1,
                                                      'host': 'host1.example.com',
                                                      'port': 8031,
                                                      'rack': 'rack1'},
                                                     {'node_id': 101,
                                                      'host': 'host2.example.com',
                                                      'port': 8032,
                                                      'rack': 'rack2'}],
                                         'controller_id': 1,
                                         'topics': [{'error': 3,
                                                     'name': 'topic1',
                                                     'internal': False,
                                                     'partitions': None}]})


def group_coordinator():
    return GroupCoordinatorV0Response.from_dict({'error': 0,
                                                 'node_id': 1,
                                                 'host': 'host1.example.com',
                                                 'port': 8031})


def group_coordinator_error():
    return GroupCoordinatorV0Response.from_dict({'error': 15,
                                                 'node_id': -1,
                                                 'host': None,
                                                 'port': -1})


def list_groups():
    return ListGroupsV0Response.from_dict({'error': 0, 'groups': [{'group_id': 'group1',
                                                                   'protocol_type': 'protocol1'}]})


def list_groups_error():
    return ListGroupsV0Response.from_dict({'error': 1, 'groups': None})


def list_offset():
    return ListOffsetV0Response.from_dict({'responses': [{'topic': 'topic1',
                                                          'partition_responses': [{'partition': 0,
                                                                                   'error': 0,
                                                                                   'offsets': [4829]},
                                                                                  {'partition': 1,
                                                                                   'error': 0,
                                                                                   'offsets': [8904]}]}]})


def list_offset_error():
    return ListOffsetV0Response.from_dict({'responses': [{'topic': 'topic1',
                                                          'partition_responses': [{'partition': 0,
                                                                                   'error': 6,
                                                                                   'offsets': None},
                                                                                  {'partition': 1,
                                                                                   'error': 0,
                                                                                   'offsets': [8904]}]}]})


def offset_fetch():
    return OffsetFetchV1Response.from_dict({'responses': [{'topic': 'topic1',
                                                           'partition_responses': [{'partition': 0,
                                                                                    'metadata': None,
                                                                                    'error': 0,
                                                                                    'offset': 4829},
                                                                                   {'partition': 1,
                                                                                    'metadata': None,
                                                                                    'error': 0,
                                                                                    'offset': 8904}]}]})


def offset_fetch_error():
    return OffsetFetchV1Response.from_dict({'responses': [{'topic': 'topic1',
                                                           'partition_responses': [{'partition': 0,
                                                                                    'error': 6,
                                                                                    'metadata': None,
                                                                                    'offset': -1},
                                                                                   {'partition': 1,
                                                                                    'metadata': None,
                                                                                    'error': 0,
                                                                                    'offset': 8904}]}]})


def offset_commit_response():
    return OffsetCommitV2Response.from_dict({'responses': [{'topic': 'topic1',
                                                            'partition_responses': [{'partition': 0,
                                                                                     'error': 0},
                                                                                    {'partition': 1,
                                                                                     'error': 16}]}]})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import division

from kafka.tools.assigner.actions import ActionModule


class ActionReorder(ActionModule):
    name = "reorder"
    helpstr = "Reelect partition leaders using replica reordering"

    def process_cluster(self):
        # Start all the leader counts at zero
        leaders = {}
        for broker in self.cluster.brokers:
            leaders[broker] = 0

        for partition in self.cluster.partitions(self.args.exclude_topics):
            # The best leader is either:
            #  1) The first replica that has 0 leaders so far
            #  2) The replica with the lowest leader ratio
            new_leader = None
            min_ratio = None
            for replica in partition.replicas:
                if leaders[replica.id] == 0:
                    new_leader = replica
                    break
            if new_leader is None:
                for replica in partition.replicas:
                    leader_ratio = leaders[replica.id] / replica.num_partitions()
                    if (min_ratio is None) or (leader_ratio < min_ratio):
                        min_ratio = leader_ratio
                        new_leader = replica

            # If the leader changed, swap the new leader with the current leader
            if new_leader != partition.replicas[0]:
                partition.swap_replica_positions(new_leader, partition.replicas[0])

            # Update the brokers hash
            leaders[new_leader.id] += 1



class BaseModel(object):
    equality_attrs = []

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError
        return not any([getattr(self, attr_name) != getattr(other, attr_name) for attr_name in self.equality_attrs])

    def __hash__(self):
        return id(self)



import abc
import six

from kafka.tools.protocol.types.sequences import Sequence


@six.add_metaclass(abc.ABCMeta)
class BaseResponse():  # pragma: no cover
    @abc.abstractproperty
    def schema(self):
        pass

    @classmethod
    def from_dict(cls, value_dict):
        return cls(Sequence(value_dict, cls.schema))

    @classmethod
    def from_bytes(cls, correlation_id, byte_array):
        seq_obj, byte_array = Sequence.decode(byte_array, schema=cls.schema)
        rv = cls(seq_obj)
        rv.correlation_id = correlation_id
        return rv

    def __init__(self, sequence_obj):
        self._response = sequence_obj

    def __hash__(self):
        return id(self)

    def __str__(self):
        return str(self._response)

    def __len__(self):
        return len(self._response)

    def __contains__(self, k):
        return k in self._response

    def __getitem__(self, k):
        return self._response[k]

    def __setitem__(self, k):
        raise NotImplementedError

    def __delitem__(self, k):
        raise NotImplementedError



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from operator import attrgetter

from kafka.tools import log
from kafka.tools.assigner.actions import ActionBalanceModule


class ActionBalanceCount(ActionBalanceModule):
    name = "count"
    helpstr = "Move the smallest partitions in the cluster to even the partition count per-broker for each replica position"

    def process_cluster(self):
        log.info("Starting partition balance by count")

        # Figure out the max RF for the cluster and sort all partition lists by size (ascending)
        max_pos = self.cluster.max_replication_factor()
        for broker in self.cluster.brokers:
            for pos in self.cluster.brokers[broker].partitions:
                self.cluster.brokers[broker].partitions[pos].sort(key=attrgetter('size'))

        # Calculate partition counts for each position first
        max_count = {}
        for pos in range(max_pos):
            # Calculate the maximum number of partitions each broker should have (floor(average))
            # We'll also track a remainder and make sure they only go 1 per broker
            pcount = 0
            for broker in self.cluster.brokers:
                if pos in self.cluster.brokers[broker].partitions:
                    pcount += self.cluster.brokers[broker].num_partitions_at_position(pos)
            max_count[pos] = [pcount / len(self.cluster.brokers), pcount % len(self.cluster.brokers)]
            log.info("Calculating ideal state for replica position {0} - max {1} partitions".format(pos, max_count[pos][0] + 1))

        # Balance partition counts for each replica position separately
        for pos in range(max_pos):
            for broker_id in self.cluster.brokers:
                broker = self.cluster.brokers[broker_id]
                # Figure out how many more partitions this broker needs
                diff = max_count[pos][0]
                if max_count[pos][1]:
                    diff += 1
                    max_count[pos][1] -= 1
                if pos in broker.partitions:
                    diff -= broker.num_partitions_at_position(pos)

                if diff > 0:
                    log.debug("Moving {0} partitions to broker {1}".format(diff, broker_id))

                    # Iterate through the largest brokers to find diff partitions to move to this broker
                    for source_id in self.cluster.brokers:
                        source = self.cluster.brokers[source_id]
                        if diff == 0:
                            break
                        if pos not in source.partitions:
                            continue

                        iterlist = list(source.partitions[pos])
                        for partition in iterlist:
                            # If we have moved enough partitions from this broker, exit out of the inner loop
                            if (source.num_partitions_at_position(pos) < max_count[pos][0]) or (diff == 0):
                                break

                            # Skip topics that are being excluded
                            if partition.topic.name in self.args.exclude_topics:
                                continue

                            # If the partition is already on the target, swap positions only if it makes the balance better
                            if broker in partition.replicas:
                                other_pos = partition.replicas.index(broker)
                                if (other_pos in source.partitions) and (source.num_partitions_at_position(other_pos) < max_count[other_pos][0]):
                                    partition.swap_replica_positions(source, broker)
                            else:
                                partition.swap_replicas(source, broker)
                                diff -= 1

                    log.debug("Finish broker {0} with {1} partitions".format(broker_id, broker.num_partitions_at_position(pos)))
                elif diff < 0:
                    log.debug("Moving {0} partitions off broker {1}".format(-diff, broker_id))

                    # Iterate through the smallest brokers to find diff partitions to move off this broker
                    for target_id in self.cluster.brokers:
                        target = self.cluster.brokers[target_id]
                        if diff == 0:
                            break
                        if (pos in target.partitions) and (target.num_partitions_at_position(pos) > (max_count[pos][0] + 1)):
                            continue

                        iterlist = list(broker.partitions[pos])
                        for partition in iterlist:
                            # If we have moved enough partitions to this broker, exit out of the inner loop
                            if ((pos in target.partitions) and (target.num_partitions_at_position(pos) >= max_count[pos][0])) or (diff == 0):
                                break
                            # Skip partitions that are already on the target broker or are being excluded
                            if (target in partition.replicas) or (partition.topic.name in self.args.exclude_topics):
                                continue

                            partition.swap_replicas(broker, target)
                            diff += 1

                    log.debug("Finish broker {0} with {1} partitions".format(broker, broker.num_partitions_at_position(pos)))
                else:
                    log.debug("Skipping broker {0} which has {1} partitions".format(broker, broker.num_partitions_at_position(pos)))
                    continue



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


# Base class that all sizers inherit from
class SizerModule(object):
    name = ""
    helpstr = ""

    ##########################################################################################################
    # OVERRIDE, BUT CALL ME - These are all the things you can/must override in your child class
    def __init__(self, args, cluster):
        self.args = args
        self.cluster = cluster

        self.properties = {}
        for pair in self.args.property:
            k, v = pair.split('=', 2)
            self.properties[k] = v

    def get_partition_sizes(self):
        pass

    ##########################################################################################################
    # MODULE BASICS - You probably shouldn't modify past here unless you're changing the basic way all modules work



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.responses import BaseResponse


class DescribeGroupsV0Response(BaseResponse):
    schema = [
        {'name': 'groups',
         'type': 'array',
         'item_type': [
             {'name': 'error', 'type': 'int16'},
             {'name': 'group_id', 'type': 'string'},
             {'name': 'state', 'type': 'string'},
             {'name': 'protocol_type', 'type': 'string'},
             {'name': 'protocol', 'type': 'string'},
             {'name': 'members',
              'type': 'array',
              'item_type': [
                  {'name': 'member_id', 'type': 'string'},
                  {'name': 'client_id', 'type': 'string'},
                  {'name': 'client_host', 'type': 'string'},
                  {'name': 'member_metadata', 'type': 'bytes'},
                  {'name': 'member_assignment', 'type': 'bytes'},
              ]},
         ]},
    ]



try:  # pragma: no cover
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:  # pragma: no cover
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)



import sys
import unittest

from kafka.tools.protocol.arguments import set_up_arguments


class ArgumentTests(unittest.TestCase):
    def test_get_arguments_none(self):
        sys.argv = ['kafka-protocol']
        args = set_up_arguments()

        # These are the current defaults, but they're not important as to the specific values
        # We're more concerned that they just exist
        assert args.broker == 'localhost'
        assert args.port == 9092
        assert not args.tls



import json
import unittest

from mock import call, patch, ANY
from subprocess import PIPE
from testfixtures import compare
from testfixtures.popen import MockPopen

from kafka.tools.exceptions import ReassignmentFailedException
from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.assigner.models.reassignment import Reassignment
from kafka.tools.assigner.plugins import PluginModule


class ReassignmentTests(unittest.TestCase):
    def setUp(self):
        self.topic = Topic('testTopic', 10)
        self.broker = Broker('brokerhost1.example.com', id=1)
        for i in range(10):
            self.topic.partitions[i].replicas = [self.broker]
        self.reassignment = Reassignment(self.topic.partitions, pause_time=0)
        self.null_plugin = PluginModule()

    def test_reassignment_create(self):
        assert self.reassignment is not None

    def test_reassignment_dict(self):
        t_repr = self.reassignment.dict_for_reassignment()
        expect_repr = {'version': 1, 'partitions': []}
        for i in range(10):
            expect_repr['partitions'].append({'topic': 'testTopic', 'partition': i, 'replicas': [1]})
        assert t_repr == expect_repr

    def test_reassignment_repr(self):
        t_repr = json.loads(repr(self.reassignment))
        expect_repr = {'version': 1, 'partitions': []}
        for i in range(10):
            expect_repr['partitions'].append({'topic': 'testTopic', 'partition': i, 'replicas': [1]})
        assert t_repr == expect_repr

    @patch.object(Reassignment, '_execute')
    def test_reassignment_execute_real(self, mock_exec):
        self.reassignment.execute(1, 1, 'zkconnect', '/path/to/tools', plugins=[self.null_plugin], dry_run=False)
        mock_exec.assert_called_once_with(1, 1, 'zkconnect', '/path/to/tools')

    @patch.object(Reassignment, '_execute')
    def test_reassignment_execute_dryrun(self, mock_exec):
        self.reassignment.execute(1, 1, 'zkconnect', '/path/to/tools', plugins=[self.null_plugin], dry_run=True)
        mock_exec.assert_not_called()

    @patch('kafka.tools.assigner.models.reassignment.subprocess.Popen', new_callable=MockPopen)
    @patch.object(Reassignment, 'check_completion')
    def test_reassignment_internal_execute(self, mock_check, mock_popen):
        mock_popen.set_default()
        mock_check.side_effect = [10, 5, 0]

        self.reassignment._execute(1, 1, 'zkconnect', '/path/to/tools')

        compare([call.Popen(['/path/to/tools/kafka-reassign-partitions.sh', '--execute', '--zookeeper', 'zkconnect', '--reassignment-json-file', ANY],
                            stderr=ANY, stdout=ANY),
                 call.Popen_instance.wait()], mock_popen.mock.method_calls)
        assert len(mock_check.mock_calls) == 3

    @patch('kafka.tools.assigner.models.reassignment.subprocess.Popen', new_callable=MockPopen)
    def test_check_completion_calls_tool(self, mock_popen):
        cmd_stdout = ("Status of partition reassignment:\n"
                      "Reassignment of partition [testTopic,0] completed successfully\n"
                      "Reassignment of partition [testTopic,1] completed successfully\n"
                      "Reassignment of partition [testTopic,2] completed successfully\n"
                      "Reassignment of partition [testTopic,3] completed successfully\n")
        mock_popen.set_default(stdout=cmd_stdout.encode('utf-8'))
        self.reassignment.check_completion('zkconnect', '/path/to/tools', 'assignfilename')
        compare([call.Popen(['/path/to/tools/kafka-reassign-partitions.sh', '--verify', '--zookeeper', 'zkconnect', '--reassignment-json-file',
                             'assignfilename'], stderr=ANY, stdout=PIPE)],
                mock_popen.mock.method_calls)

    def test_verify_regex_bad(self):
        assert self.reassignment.status_re.match("Status of partition reassignment:\n") is None

    def test_verify_regex_failed(self):
        match_obj = self.reassignment.status_re.match("Reassignment of partition [testTopic,1] failed\n")
        assert match_obj is not None
        assert match_obj.group(1) == "failed"

    def test_verify_regex_progress(self):
        match_obj = self.reassignment.status_re.match("Reassignment of partition [testTopic,1] still in progress\n")
        assert match_obj is not None
        assert match_obj.group(1) == "still in progress"

    def test_verify_regex_success(self):
        match_obj = self.reassignment.status_re.match("Reassignment of partition [testTopic,1] completed successfully\n")
        assert match_obj is not None
        assert match_obj.group(1) == "completed successfully"

    def test_process_verify_match_failed(self):
        count = self.reassignment.process_verify_match("Reassignment of partition [testTopic,1] failed\n")
        assert count == -1

    def test_process_verify_match_progress(self):
        count = self.reassignment.process_verify_match("Reassignment of partition [testTopic,1] still in progress\n")
        assert count == 1

    def test_process_verify_match_other(self):
        count = self.reassignment.process_verify_match("Reassignment of partition [testTopic,1] completed successfully\n")
        assert count == 0

    @patch('kafka.tools.assigner.models.reassignment.subprocess.Popen', new_callable=MockPopen)
    def test_check_completion_failed(self, mock_popen):
        cmd_stdout = ("Status of partition reassignment:\n"
                      "ERROR: Assigned replicas (1,2) don't match the list of replicas for reassignment (1,2,3) for partition [testTopic,1]\n"
                      "Reassignment of partition [testTopic,0] completed successfully\n"
                      "Reassignment of partition [testTopic,1] failed\n"
                      "Reassignment of partition [testTopic,2] still in progress\n"
                      "Reassignment of partition [testTopic,3] completed successfully\n")
        mock_popen.set_default(stdout=cmd_stdout.encode('utf-8'))
        self.assertRaises(ReassignmentFailedException, self.reassignment.check_completion, 'zkconnect', '/path/to/tools', 'assignfilename')

    @patch('kafka.tools.assigner.models.reassignment.subprocess.Popen', new_callable=MockPopen)
    def test_check_completion_success(self, mock_popen):
        cmd_stdout = ("Status of partition reassignment:\n"
                      "Reassignment of partition [testTopic,0] completed successfully\n"
                      "Reassignment of partition [testTopic,1] completed successfully\n"
                      "Reassignment of partition [testTopic,2] completed successfully\n"
                      "Reassignment of partition [testTopic,3] completed successfully\n")
        mock_popen.set_default(stdout=cmd_stdout.encode('utf-8'))
        assert self.reassignment.check_completion('zkconnect', '/path/to/tools', 'assignfilename') == 0

    @patch('kafka.tools.assigner.models.reassignment.subprocess.Popen', new_callable=MockPopen)
    def test_check_completion_progress(self, mock_popen):
        cmd_stdout = ("Status of partition reassignment:\n"
                      "Reassignment of partition [testTopic,0] completed successfully\n"
                      "Reassignment of partition [testTopic,1] still in progress\n"
                      "Reassignment of partition [testTopic,2] completed successfully\n"
                      "Reassignment of partition [testTopic,3] completed successfully\n")
        mock_popen.set_default(stdout=cmd_stdout.encode('utf-8'))
        assert self.reassignment.check_completion('zkconnect', '/path/to/tools', 'assignfilename') == 1



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import division

from kafka.tools.exceptions import ReplicaNotFoundException
from kafka.tools.models import BaseModel


class Partition(BaseModel):
    equality_attrs = ['topic', 'num']

    def __init__(self, topic, num):
        self.topic = topic
        self.num = num
        self.leader = None
        self.replicas = []
        self.size = 0
        self.scaled_size = 0

    # Shallow copy - do not copy replica list (zero length)
    def copy(self):
        newpartition = Partition(self.topic, self.num)
        newpartition.size = self.size
        return newpartition

    # Set the current size of this partition iff it is larger than the currently known size
    def set_size(self, size):
        if size > self.size:
            self.size = size
            self.scaled_size = (self.topic.cluster.retention / self.topic.retention) * self.size

    def dict_for_reassignment(self):
        return {"topic": self.topic.name, "partition": self.num, "replicas": [broker.id for broker in self.replicas]}

    def dict_for_replica_election(self):
        return {"topic": self.topic.name, "partition": self.num}

    # Given a broker, add it to this partition as a replica
    # If position is not specified, default to the end of the replica list
    def add_replica(self, broker, position=-1):
        if position < 0:
            position = len(self.replicas)
        self._add_broker_partition(position, broker)
        self.replicas.insert(position, broker)

    # Remove the specified broker from the replica list of this partition
    # If the replica does not exist on this partition, throw an exception
    def remove_replica(self, broker):
        try:
            position = self.replicas.index(broker)
        except ValueError:
            raise ReplicaNotFoundException

        broker.partitions[position].remove(self)
        self.replicas.remove(broker)

    def delete_replicas(self, target_count):
        """
        Assure that the partition has only the specified number of replicas, deleting any extras

        Args:
            target_count (int): the maximum number of replicas to retain
        """
        while len(self.replicas) > target_count:
            self.remove_replica(self.replicas[-1])

    def add_or_update_replica(self, position, new_broker):
        """
        Given a position and a broker, make sure the broker is in the replica
        set for this partition at the given position.

        Args:
            position (int): The position in the replica list for the new broker
            new_broker (Broker): The broker that should now be at that position in the replica set
        """
        if len(self.replicas) > position:
            if self.replicas[position] == new_broker:
                # No change in the replica at this position
                return
            else:
                # New replica at this position. Swap it in
                self.swap_replicas(self.replicas[position], new_broker)
        else:
            # No replica yet at this position. Add it
            self.add_replica(new_broker, position=position)

    # Remove one broker from the replica list and replace it at the same position with another
    # This just calls add_replica and remove_replica, but we do it a lot
    def swap_replicas(self, remove_broker, add_broker):
        try:
            position = self.replicas.index(remove_broker)
        except ValueError:
            raise ReplicaNotFoundException

        self.remove_replica(remove_broker)
        self.add_replica(add_broker, position)

    # Given two brokers that appear in the replica list, swap their positions (making sure to adjust the broker objects as well)
    # If either replica is not in the list, throw an exception
    def swap_replica_positions(self, broker1, broker2):
        try:
            p1 = self.replicas.index(broker1)
            p2 = self.replicas.index(broker2)
        except ValueError:
            raise ReplicaNotFoundException

        # First, change the position of this partition on the first broker object to p2
        self._remove_broker_partition(broker1)
        self._add_broker_partition(p2, broker1)

        # Then change the position of this partition on the second broker object to p1
        self._remove_broker_partition(broker2)
        self._add_broker_partition(p1, broker2)

        # Last, swap the replica positons on this partition object
        self.replicas[p1] = broker2
        self.replicas[p2] = broker1

    # Helper function to add a partition to a broker
    # This should never be called - please use the add_replica method
    def _add_broker_partition(self, pos, broker):
        if pos not in broker.partitions:
            broker.partitions[pos] = [self]
        else:
            broker.partitions[pos].append(self)

    # Helper function to remove a partition from a broker
    # This should never be called - please use the remove_replica method
    def _remove_broker_partition(self, broker):
        pos = self.replicas.index(broker)
        broker.partitions[pos].remove(self)



import six
import unittest

from kafka.tools.protocol.errors import error_short, error_long


class ErrorTests(unittest.TestCase):
    def test_error_short(self):
        assert error_short(3) == 'UNKNOWN_TOPIC_OR_PARTITION'

    def test_error_short_unknown(self):
        assert error_short(99) == 'NOSUCHERROR'

    def test_error_long(self):
        assert isinstance(error_long(3), six.string_types)

    def test_error_long_unknown(self):
        assert isinstance(error_long(99), six.string_types)



import unittest

from kafka.tools.protocol.types.string import String


class StringTests(unittest.TestCase):
    def test_create(self):
        val = String('foo')
        assert val.value() == 'foo'

    def test_create_null(self):
        val = String(None)
        assert val.value() is None

    def test_encode_bad_type(self):
        self.assertRaises(TypeError, String, 123)

    def test_decode(self):
        (val, rest) = String.decode(b'\x00\x03foo')
        assert isinstance(val, String)
        assert val.value() == 'foo'
        assert rest == b''

    def test_decode_null(self):
        (val, rest) = String.decode(b'\xff\xff')
        assert isinstance(val, String)
        assert val.value() is None
        assert rest == b''

    def test_decode_remainder(self):
        (val, rest) = String.decode(b'\x00\x03fooabc')
        assert val.value() == 'foo'
        assert rest == b'abc'

    def test_decode_nosize(self):
        self.assertRaises(ValueError, String.decode, b'')

    def test_decode_underflow(self):
        self.assertRaises(ValueError, String.decode, b'\x00\x03fo')

    def test_encode(self):
        val = String('foo')
        assert val.encode() == b'\x00\x03foo'

    def test_encode_null(self):
        val = String(None)
        assert val.encode() == b'\xff\xff'

    def test_str(self):
        val = String('foo')
        assert str(val) == "foo (string)"
        assert repr(val) == "<String length=3>"



import unittest

from kafka.tools.protocol.types.boolean import Boolean


class BooleanTests(unittest.TestCase):
    def test_create_true(self):
        val = Boolean(True)
        assert val.value()

    def test_create_one(self):
        val = Boolean(1)
        assert val.value()

    def test_create_false(self):
        val = Boolean(False)
        assert not val.value()

    def test_create_zero(self):
        val = Boolean(0)
        assert not val.value()

    def test_decode_true(self):
        (val, rest) = Boolean.decode(b'\x01')
        assert isinstance(val, Boolean)
        assert val.value()
        assert rest == b''

    def test_decode_false(self):
        (val, rest) = Boolean.decode(b'\x00')
        assert isinstance(val, Boolean)
        assert not val.value()
        assert rest == b''

    def test_decode_remainder(self):
        (val, rest) = Boolean.decode(b'\x01abc')
        assert val.value()
        assert rest == b'abc'

    def test_decode_underflow(self):
        self.assertRaises(ValueError, Boolean.decode, b'')

    def test_encode_true(self):
        val = Boolean(True)
        assert val.encode() == b'\x01'

    def test_encode_false(self):
        val = Boolean(False)
        assert val.encode() == b'\x00'

    def test_boolean_context(self):
        val = Boolean(True)
        val2 = Boolean(False)
        assert val
        assert not val2

    def test_str(self):
        val = Boolean(True)
        assert str(val) == "True (boolean)"
        assert repr(val) == "<Boolean True>"
